#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **command runners** (i.e., callables running external commands as
subprocesses of the active Python process).

This private submodule is *not* intended for importation by downstream callers.

Environment Variables
---------------------
Environment variables may be "passed" to external commands in various ways.
With respect to test-specific command execution, the ideal approach is to (in
order):

#. Require the ``monkeypatch : 'pytest.MonkeyPatch'`` fixture.
#. Call the ``monkeypatch.setenv({{env_name}}, {{env_value}})`` method as needed
   to temporarily set all requisite environment variables immediately *before*
   calling a command runner defined by this submodule.

Command Words Arguments
-----------------------
Most runners accept a mandatory ``command_words`` parameter, a list of one or
more shell words comprising this command whose:

* Mandatory first item is either:

  * This command's absolute or relative path.
  * This command's basename, in which case the first command with that basename
    in the current ``${PATH}`` environment variable will be run. If no such
    command is found, an exception is raised.

* Optional subsequent items are this command's arguments (in order). Note that
  these arguments are passed as is to a low-level system call rather than
  intprereted by a high-level shell (e.g., ``/bin/sh`` on POSIX-compatible
  platforms) and hence should *not* be shell-quoted. Indeed, shell quoting
  these arguments is likely to result in erroneous command behaviour. The
  principal exception to this heuristic are **GNU-style long value options**
  (i.e., ``--``-prefixed options accepting a ``=``-delimited value), whose
  values should either be:

  * Passed as a separate argument *without* being shell-quoted.
  * Concatenated to the current ``--``-prefixed argument delimited by ``=`` and
    shell-quoted.

``Popen()`` Keyword Arguments
-----------------------------
Most runners accept the same optional keyword arguments accepted by the
:meth:`subprocess.Popen.__init__` constructor, including:

* ``cwd``, the absolute path of the current working directory (CWD) from which
  this command is to be run. Defaults to the current CWD. **Unfortunately, note
  that this keyword argument appears to be erroneously ignored on numerous
  platforms** (e.g., Windows). For safety, a context manager temporarily
  changing the current directory should typically be leveraged instead.
* ``timeout``, the maximum number of milliseconds this command is to be run
  for. Commands with execution time exceeding this timeout will be mercifully
  killed. Defaults to :data:`None`, in which case this command is run
  indefinitely.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import (
    # Iterable,
    Mapping,
    Optional,
    Tuple,
)
from beartype._data.typing.datatyping import CommandWords
from collections.abc import (
    Iterable as IterableABC,
    Mapping as MappingABC,
)
from os import environ
from subprocess import (
    CalledProcessError,
    TimeoutExpired,
    call as subprocess_call,
    check_output as subprocess_check_output,
    run as subprocess_run,
)

# ....................{ GLOBALS                            }....................
BUFFER_SIZE_DEFAULT = -1
'''
Default subprocess buffer size for the current platform (synonymous with the
current :data:`io.DEFAULT_BUFFER_SIZE`) suitable for passing as the ``bufsize``
parameter accepted by :meth:`subprocess.Popen.__init__` method.
'''


BUFFER_SIZE_NONE = 0
'''
Unbuffered subprocess buffer size suitable for passing as the ``bufsize``
parameter accepted by :meth:`subprocess.Popen.__init__` method.

Both reading from and writing to an unbuffered subprocess is guaranteed to
perform exactly one system call (``read()`` and ``write()``, respectively) and
can return short (i.e., produce less bytes than the requested number of bytes).
'''


BUFFER_SIZE_LINE = 1
'''
Line-buffered subprocess buffer size suitable for passing as the ``bufsize``
parameter accepted by :meth:`subprocess.Popen.__init__` method.

Reading from a line-buffered subprocess is guaranteed to block until the
subprocess emits a newline, at which point all output emitted between that
newline inclusive and the prior newline exclusive is consumed.
'''

# ....................{ PRIVATE ~ hints                    }....................
_HINT_POPEN_KWARGS = Mapping[str, object]
'''
PEP-compliant type hint matching the return value of the private
:func:`_init_popen_kwargs` function.
'''


_HINT_POPEN_KWARGS_OPTIONAL = Optional[_HINT_POPEN_KWARGS]
'''
PEP-compliant type hint matching the ``popen_kwargs`` parameter passed to most
callables declared by this submodule.
'''

# ....................{ RUNNERS ~ command                  }....................
#FIXME: Unit test us up, please.
def run_command_forward_stderr_return_stdout(
    # Mandatory parameters.
    command_words: CommandWords,

    # Optional parameters.
    popen_kwargs: _HINT_POPEN_KWARGS_OPTIONAL = None,
) -> str:
    '''
    Run the passed command as a subprocess of the active Python process, raising
    an exception on subprocess failure while both forwarding all standard error
    of this subprocess to the standard error file handle of the active Python
    process *and* capturing and returning all standard output of this
    subprocess.

    This exception contains the exit status of this subprocess.

    Parameters
    ----------
    command_words : CommandWords
        Iterable of one or more shell words comprising this command.
    popen_kwargs : _HINT_POPEN_KWARGS_OPTIONAL
        Dictionary of all keyword arguments to be passed to the
        :meth:`subprocess.Popen.__init__` method. Defaults to :data:`None`, in
        which case the empty dictionary is assumed.

    Returns
    -------
    str
        All standard output captured from this subprocess, stripped of all
        trailing newlines (as under most POSIX shells) *and* decoded with the
        current locale's preferred encoding (e.g., UTF-8).

    Raises
    ------
    CalledProcessError
        If the subprocess running this command report non-zero exit status.
    '''

    # Sanitize these arguments.
    popen_kwargs = _init_popen_kwargs(command_words, popen_kwargs)

    # Capture this command's stdout, raising an exception on command failure
    # (including failure due to an expired timeout).
    command_stdout = subprocess_check_output(command_words, **popen_kwargs)

    # Return this stdout, stripped of all trailing newlines.
    return command_stdout.rstrip('\n')


#FIXME: Unit test us up, please.
def run_command_forward_output(
    # Mandatory parameters.
    command_words: CommandWords,

    # Optional parameters.
    popen_kwargs: _HINT_POPEN_KWARGS_OPTIONAL = None,
) -> None:
    '''
    Run the passed command as a subprocess of the active Python process,
    raising an exception on subprocess failure while forwarding all standard
    output and error output by this subprocess to the standard output and error
    file handles of the active Python process.

    This exception contains the exit status of this subprocess.

    Parameters
    ----------
    command_words : CommandWords
        Iterable of one or more shell words comprising this command.
    popen_kwargs : _HINT_POPEN_KWARGS_OPTIONAL
        Dictionary of all keyword arguments to be passed to the
        :meth:`subprocess.Popen.__init__` method. Defaults to :data:`None`, in
        which case the empty dictionary is assumed.

    Raises
    ------
    CalledProcessError
        If the subprocess running this command report non-zero exit status.
    '''

    # Defer test-specific imports.
    from beartype_test._util.command.pytcmdexit import is_failure

    # 0-based exit status reported by running this command.
    exit_status = run_command_forward_output_return_status(
        command_words=command_words, popen_kwargs=popen_kwargs)

    # If this command failed, raising an exception on command failure. For
    # reusability, we reimplement the subprocess.check_call() function here
    # rather than explicitly call that function. The latter approach would
    # require duplicating logic between this and the
    # run_command_forward_output_return_status() runner called above.
    if is_failure(exit_status):
        raise CalledProcessError(exit_status, command_words)
    # Else, this command succeeded.


#FIXME: Unit test us up, please.
def run_command_forward_output_return_status(
    # Mandatory parameters.
    command_words: CommandWords,

    # Optional parameters.
    popen_kwargs: _HINT_POPEN_KWARGS_OPTIONAL = None
) -> int:
    '''
    Run the passed command as a subprocess of the active Python process,
    returning only the exit status of this subprocess while forwarding all
    standard output and error output by this subprocess to the standard output
    and error file handles of the active Python process.

    Caveats
    -------
    **This function raises no exceptions on subprocess failure.** To do so,
    consider calling the :func:`run_command_forward_output` function instead.

    Parameters
    ----------
    command_words : CommandWords
        Iterable of one or more shell words comprising this command.
    popen_kwargs : _HINT_POPEN_KWARGS_OPTIONAL
        Dictionary of all keyword arguments to be passed to the
        :meth:`subprocess.Popen.__init__` method. Defaults to :data:`None`, in
        which case the empty dictionary is assumed.

    Returns
    -------
    int
        Exit status returned by this subprocess.
    '''

    # Defer test-specific imports.
    from beartype_test._util.command.pytcmdexit import FAILURE_DEFAULT

    # Sanitize these arguments.
    popen_kwargs = _init_popen_kwargs(command_words, popen_kwargs)

    # Run this command *WITHOUT* raising an exception on command failure.
    try:
        exit_status = subprocess_call(command_words, **popen_kwargs)
    # If this command failed to halt before triggering a timeout, the "timeout"
    # keyword argument was passed *AND* this command has effectively failed.
    # Since the prior call has already guaranteeably terminated this command,
    # this exception is safely convertible into default failure exit status.
    except TimeoutExpired:
        exit_status = FAILURE_DEFAULT

    # Return this exit status.
    return exit_status


def run_command_return_stdout_stderr(
    # Mandatory parameters.
    command_words: CommandWords,

    # Optional parameters.
    popen_kwargs: _HINT_POPEN_KWARGS_OPTIONAL = None,
) -> Tuple[str, str]:
    '''
    Run the passed command as a subprocess of the active Python process, raising
    an exception on subprocess failure while capturing and returning both all
    standard output and error of this subprocess.

    This exception contains the exit status of this subprocess.

    Parameters
    ----------
    command_words : CommandWords
        Iterable of one or more shell words comprising this command.
    popen_kwargs : _HINT_POPEN_KWARGS_OPTIONAL
        Dictionary of all keyword arguments to be passed to the
        :meth:`subprocess.Popen.__init__` method. Defaults to :data:`None`, in
        which case the empty dictionary is assumed.

    Returns
    -------
    Tuple[str, str]
        All standard output and error (in that order) captured from this
        subprocess, stripped of all trailing newlines (as under most POSIX
        shells) *and* decoded with the current locale's preferred encoding
        (e.g., UTF-8).

    Raises
    ------
    CalledProcessError
        If the subprocess running this command reports non-zero exit status.
    '''

    # If these keyword arguments are empty, default to the empty dictionary
    # *BEFORE* setting dictionary keys below.
    if popen_kwargs is None:
        popen_kwargs = {}
    # Else, these keyword arguments are non-empty.

    # In either case, these keyword arguments are now a dictionary. Assert this.
    assert isinstance(popen_kwargs, MappingABC), (
        f'{repr(popen_kwargs)} not mapping.')

    # Enable capturing of both standard output and error.
    popen_kwargs['capture_output'] = True

    # Enable raising of a "CalledProcessError" exception when this command
    # reports non-zero exit status.
    popen_kwargs['check'] = True

    # Sanitize these arguments.
    popen_kwargs = _init_popen_kwargs(command_words, popen_kwargs)

    # "subprocess.CompletedProcess" object encapsulating the result of running
    # this command.
    command_result = subprocess_run(command_words, **popen_kwargs)

    # Standard output and error emitted by this command, stripped of all
    # trailing newlines.
    command_stdout = command_result.stdout.rstrip('\n')
    command_stderr = command_result.stderr.rstrip('\n')

    # Return this standard output and error.
    return command_stdout, command_stderr

# ....................{ PRIVATE ~ constants                }....................
_INIT_POPEN_KWARGS_POPEN_KWARGS_NAMES_CLOSE_FDS_CONFLICTING = frozenset((
    'stdin', 'stdout', 'stderr', 'close_fds'))
'''
Frozen set of the names of all keyword parameters which if passed would prevent
the :func:`_init_popen_kwargs` function from safely defaulting the ``close_fds``
parameter to false under **vanilla Microsoft Windows** (i.e., *not* running the
Cygwin POSIX compatibility layer).

See Also
----------
:func:`_init_popen_kwargs`
    Further details.
'''

# ....................{ PRIVATE ~ initializers             }....................
def _init_popen_kwargs(
    # Mandatory parameters.
    command_words: CommandWords,

    # Optional parameters.
    popen_kwargs: _HINT_POPEN_KWARGS_OPTIONAL = None
) -> _HINT_POPEN_KWARGS:
    '''
    Sanitized dictionary of all keyword arguments to pass to the
    :class:`subprocess.Popen` callable when running the command specified by
    the passed shell words with the passed user-defined keyword arguments.

    Caveats
    -------
    If the current platform is vanilla Windows *and* none of the ``stdin``,
    ``stdout``, ``stderr``, or ``close_fds`` parameters are passed, this
    function defaults the ``close_fds`` parameter if unpassed to :data:`False`.
    Doing so causes this command to inherit all file handles (including stdin,
    stdout, and stderr) from the active Python process. Note that the
    :class:`subprocess.Popen` docstring insists that:

        On Windows, if ``close_fds`` is :data:`True` then no handles will be
        inherited by the child process.

    The child process will then open new file handles for stdin, stdout, and
    stderr. If the current terminal is a Windows Console, the underlying
    terminal devices and hence file handles will remain the same, in which case
    this is *not* an issue. If the current terminal is Cygwin-based (e.g.,
    MinTTY), however, the underlying terminal devices and hence file handles
    will differ, in which case this behaviour prevents interaction between the
    current shell and the vanilla Windows command to be run below. In
    particular, all output from this command will be squelched.

    If at least one of stdin, stdout, or stderr are redirected to a blocking
    pipe, setting ``close_fds`` to :data:`False` can induce deadlocks under
    certain edge-case scenarios. Since all such file handles default to
    :data:`None` and hence are *not* redirected in this case, ``close_fds`` may
    be safely set to :data:`False`.

    On all other platforms, if ``close_fds`` is :data:`True`, no file handles
    *except* stdin, stdout, and stderr will be inherited by the child process.
    This function fundamentally differs in subtle (and only slightly documented
    ways) between vanilla Windows and all other platforms. These discrepancies
    appear to be harmful but probably unavoidable, given the philosophical gulf
    between vanilla Windows and all other platforms.

    Parameters
    ----------
    command_words : CommandWords
        Iterable of one or more shell words comprising this command.
    popen_kwargs : _HINT_POPEN_KWARGS_OPTIONAL
        Dictionary of all keyword arguments to be passed to the
        :meth:`subprocess.Popen.__init__` method. Defaults to :data:`None`, in
        which case the empty dictionary is assumed.

    Returns
    -------
    _HINT_POPEN_KWARGS
        This dictionary of keyword arguments sanitized.
    '''

    # Defer test-specific imports.
    from beartype._util.kind.maplike.utilmaptest import is_mapping_keys_any
    from beartype._util.os.utilostest import is_os_windows_vanilla

    # If these keyword arguments are empty, default to the empty dictionary
    # *BEFORE* validating these arguments as a dictionary below.
    if popen_kwargs is None:
        popen_kwargs = {}
    # Else, these keyword arguments are non-empty.
    #
    # In either case, these keyword arguments are now a dictionary.

    # Validate these parameters *AFTER* defaulting them above if needed.
    assert isinstance(command_words, IterableABC), (
        f'{repr(command_words)} not iterable.')
    assert bool(command_words), '"command_words" empty.'
    assert isinstance(popen_kwargs, MappingABC), (
        f'{repr(popen_kwargs)} not mapping.')

    #FIXME: Uncomment if we ever feel like implementing this additional
    #validation. For the moment, we simply let lower-level functionality in the
    #stdlib do the dirty work for us. :p
    # If the first shell word is this list is unrunnable, raise an exception.
    # die_unless_command(command_words[0])

    # Log the command to be run before doing so.
    # log_debug('Running command: %s', ' '.join(command_words))

    # If this is vanilla Windows *AND* the caller passed no keyword arguments
    # that would prevent us from safely defaulting the "close_fds" parameter to
    # false, sanitize that parameter to false.
    if is_os_windows_vanilla() and not is_mapping_keys_any(
        mapping=popen_kwargs,
        keys=_INIT_POPEN_KWARGS_POPEN_KWARGS_NAMES_CLOSE_FDS_CONFLICTING,
    ):
        popen_kwargs['close_fds'] = False

    # Isolate the current set of environment variables to this command,
    # preventing concurrent changes in these variables in the active process
    # from affecting this command's subprocess.
    popen_kwargs['env'] = environ.copy()

    # Decode command output with the current locale's preferred encoding.
    popen_kwargs['universal_newlines'] = True

    # Return these keyword arguments.
    return popen_kwargs



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Public beartype decorator.**

This private submodule defines the core :func:`beartype` decorator, which the
:mod:`beartype.__init__` submodule then imports for importation as the public
:mod:`beartype.beartype` decorator by downstream callers -- completing the
virtuous cycle of code life.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ TODO                               }....................
# All "FIXME:" comments for this submodule reside in this package's "__init__"
# submodule to improve maintainability and readability here.

# ....................{ IMPORTS                            }....................
from beartype.typing import Optional
from beartype._conf.confcommon import BEARTYPE_CONF_DEFAULT
from beartype._conf.confmain import BeartypeConf
from beartype._data.typing.datatyping import (
    BeartypeReturn,
    BeartypeableT,
)
from beartype._util.py.utilpyinterpreter import is_python_optimized
from collections.abc import Callable

# Intentionally import the standard mypy-friendly @typing.overload decorator
# rather than a possibly mypy-unfriendly @beartype.typing.overload decorator --
# which, in any case, would be needlessly inefficient and thus bad.
from typing import overload

# ....................{ OVERLOADS                          }....................
# Declare PEP 484-compliant overloads to avoid breaking downstream code
# statically type-checked by a static type checker (e.g., mypy). The concrete
# @beartype decorator declared below is permissively annotated as returning a
# union of multiple types desynchronized from the types of the passed arguments
# and thus fails to accurately convey the actual public API of that decorator.
# See also:
#     https://www.python.org/dev/peps/pep-0484/#function-method-overloading
#
# Note that the "Callable[[BeartypeableT], BeartypeableT]" type hint should
# ideally instead be a reference to our "BeartypeConfedDecorator" type hint.
# Indeed, it used to be. Unfortunately, a significant regression in mypy
# required us to inline that type hint away. See also this issue:
#     https://github.com/beartype/beartype/issues/332
@overload  # type: ignore[misc,no-overload-impl]
def beartype(obj: BeartypeableT) -> BeartypeableT: ...
@overload
def beartype(*, conf: BeartypeConf) -> (
    Callable[[BeartypeableT], BeartypeableT]): ...

# ....................{ DECORATORS                         }....................
# If the active Python interpreter is optimized either at process-invocation
# time (e.g., by the user passing one or more "-O" command-line options *OR*
# setting the '${PYTHONOPTIMIZE}" environment variable to a positive integer
# when the active Python interpreter was forked) *OR* after process-invocation
# time (e.g., by the user setting the '${PYTHONOPTIMIZE}" environment variable
# to a positive integer in an interactive REPL), then unconditionally disable
# @beartype-based type-checking across the entire codebase by reducing the
# @beartype decorator to the identity decorator.
#
# Ideally, this would have been implemented at the top rather than bottom of
# this submodule as a conditional resembling:
#     if __debug__:
#         def beartype(func: CallableTypes) -> CallableTypes:
#             return func
#         return
#
# Tragically, Python fails to support module-scoped "return" statements. *sigh*
if is_python_optimized():
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: Synchronize the signature of this identity decorator with the
    # non-identity decorator imported below.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    def beartype(  # type: ignore[no-redef]
        obj: Optional[BeartypeableT] = None,

        # Optional keyword-only parameters.
        *,
        conf: BeartypeConf = BEARTYPE_CONF_DEFAULT,
    ) -> BeartypeReturn:
        # If passed an object to be decorated, this decorator is in decoration
        # rather than configuration mode. In this case, silently reduce to a
        # noop by returning this object as is unmodified.
        #
        # Note that this is the common case.
        if obj is not None:
            return obj
        # Else, this decorator was passed *NO* object to be decorated. In this
        # case, this decorator is in configuration rather than decoration mode.

        # Pretend to configure this decorator with this configuration by instead
        # returning the identity decorator -- which trivially returns the passed
        # object unmodified. Doing so ignores this configuration in the
        # optimally efficient manner, thus complying with external user demands
        # for optimization.
        return _beartype_optimized


    def _beartype_optimized(obj: BeartypeableT) -> BeartypeableT:
        '''
        Identity decorator returning the passed **beartypeable** (i.e.,
        pure-Python callable or class) unmodified due to the active Python
        interpreter being **optimized** (e.g., by either the
        ``${PYTHONOPTIMIZE}}`` environment variable being set *or* this
        interpreter being passed one or more ``"-O"`` command-line options).

        Parameters
        ----------
        obj : BeartypeableT
            Beartypeable to be decorated.

        Returns
        -------
        BeartypeableT
            Either:

            * If the passed object is a class, this existing class embellished
              with dynamically generated type-checking.
            * If the passed object is a callable, a new callable wrapping that
              callable with dynamically generated type-checking.

        See Also
        --------
        :func:`beartype.beartype`
            Further details.
        '''

        # Silently reduce to a noop by returning this object as is unmodified.
        return obj
# Else, the active Python interpreter is in a standard runtime state. In this
# case, define the @beartype decorator in the standard way.
else:
    # This is where @beartype *REALLY* lives. Grep here for all the goods.
    from beartype._decor.decorcache import beartype  # type: ignore[no-redef]

# ....................{ DECORATORS ~ doc                   }....................
# Document the @beartype decorator with the same documentation regardless of
# which of the above implementations currently implements that decorator.
beartype.__doc__ = (
    '''
    Decorate the passed **beartypeable** (i.e., pure-Python callable or
    class) with optimal type-checking dynamically generated unique to that
    beartypeable under the passed beartype configuration.

    This decorator supports two distinct (albeit equally efficient) modes
    of operation:

    * **Decoration mode.** The caller activates this mode by passing this
      decorator a type-checkable object via the ``obj`` parameter; this
      decorator then creates and returns a new callable wrapping that object
      with optimal type-checking. Specifically:

      * If this object is a callable, this decorator creates and returns a new
        **runtime type-checker** (i.e., pure-Python function validating all
        parameters and returns of all calls to that callable against all
        PEP-compliant type hints annotating those parameters and returns). The
        type-checker returned by this decorator is:

        * Optimized uniquely for the passed callable.
        * Guaranteed to run in ``O(1)`` constant-time with negligible constant
          factors.
        * Type-check effectively instantaneously.
        * Add effectively no runtime overhead to the passed callable.

      * If the passed object is a class, this decorator iteratively applies
        itself to all annotated methods of this class by dynamically wrapping
        each such method with a runtime type-checker (as described previously).

    * **Configuration mode.** The caller activates this mode by passing this
      decorator a beartype configuration via the ``conf`` parameter; this
      decorator then creates and returns a new beartype decorator enabling that
      configuration. That decorator may then be called (in decoration mode) to
      create and return a new callable wrapping the passed type-checkable
      object with optimal type-checking configured by that configuration.

    If optimizations are enabled by the active Python interpreter (e.g., due to
    option ``-O`` passed to this interpreter), this decorator silently reduces
    to a noop.

    Parameters
    ----------
    obj : Optional[BeartypeableT]
        **Beartypeable** (i.e., pure-Python callable or class) to be decorated.
        Defaults to :data:`None`, in which case this decorator is in
        configuration rather than decoration mode. In configuration mode, this
        decorator creates and returns an efficiently cached private decorator
        that generically applies the passed beartype configuration to any
        beartypeable object passed to that decorator. Look... It just works.
    conf : BeartypeConf, optional
        **Beartype configuration** (i.e., self-caching dataclass encapsulating
        all settings configuring type-checking for the passed object). Defaults
        to ``BeartypeConf()``, the default :math:`O(1)` constant-time
        configuration.

    Returns
    -------
    BeartypeReturn
        Either:

        * If in decoration mode (i.e., ``obj`` is *not* ``None` while ``conf``
          is :data:`None`) *and*:

          * If ``obj`` is a callable, a new callable wrapping that callable
            with dynamically generated type-checking.
          * If ``obj`` is a class, this existing class embellished with
            dynamically generated type-checking.

        * If in configuration mode (i.e., ``obj`` is :data:`None` while ``conf``
          is *not* :data:`None`), a new beartype decorator enabling this
          configuration.

    Raises
    ------
    BeartypeConfException
        If the passed configuration is *not* actually a configuration (i.e.,
        instance of the :class:`BeartypeConf` class).
    BeartypeDecorHintException
        If any annotation on this callable is neither:

        * A **PEP-compliant type** (i.e., instance or class complying with a
          PEP supported by :mod:`beartype`), including:

          * :pep:`484` types (i.e., instance or class declared by the stdlib
            :mod:`typing` module).

        * A **PEP-noncompliant type** (i.e., instance or class complying with
          :mod:`beartype`-specific semantics rather than a PEP), including:

          * **Fully-qualified forward references** (i.e., strings specified as
            fully-qualified classnames).
          * **Tuple unions** (i.e., tuples containing one or more classes
            and/or forward references).
    BeartypePep563Exception
        If :pep:`563` is active for this callable and evaluating a **postponed
        annotation** (i.e., annotation whose value is a string) on this
        callable raises an exception (e.g., due to that annotation referring to
        local state no longer accessible from this deferred evaluation).
    BeartypeDecorParamNameException
        If the name of any parameter declared on this callable is prefixed by
        the reserved substring ``__beartype_``.
    BeartypeDecorWrappeeException
        If this callable is either:

        * Uncallable.
        * A class, which :mod:`beartype` currently fails to support.
        * A C-based callable (e.g., builtin, third-party C extension).
    BeartypeDecorWrapperException
        If this decorator erroneously generates a syntactically invalid wrapper
        function. This should *never* happen, but here we are, so this probably
        happened. Please submit an upstream issue with our issue tracker if you
        ever see this. (Thanks and abstruse apologies!)
    '''
)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`675`-compliant **type hint test data.**
'''

# ....................{ FIXTURES                           }....................
def hints_pep675_meta() -> 'List[HintPepMetadata]':
    '''
    Session-scoped fixture returning a list of :pep:`675`-compliant **type hint
    metadata** (i.e.,
    :class:`beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata`
    instances describing test-specific :pep:`675`-compliant sample type hints
    with metadata generically leveraged by various PEP-agnostic unit tests).
    '''

    # ..................{ IMPORTS                            }..................
    # Defer fixture-specific imports.
    from beartype._data.hint.sign.datahintsigns import (
        HintSignLiteralString)
    from beartype._util.api.standard.utiltyping import get_typing_attrs
    from beartype_test.a00_unit.data.hint.util.data_hintmetacls import (
        HintPepMetadata,
        HintPithSatisfiedMetadata,
        HintPithUnsatisfiedMetadata,
    )

    # ..................{ LOCALS                             }..................
    # List of all PEP-specific type hint metadata to be returned.
    hints_pep_meta = []

    # ..................{ FACTORIES                          }..................
    # For each PEP-specific type hint factory importable from each currently
    # importable "typing" module...
    for LiteralString in get_typing_attrs('LiteralString'):
        # ..................{ LISTS                          }..................
        # Add PEP-specific type hint metadata to this list.
        hints_pep_meta.extend((
            # ................{ PEP 675                        }................
            # Literal string type hint.
            HintPepMetadata(
                hint=LiteralString,
                pep_sign=HintSignLiteralString,
                piths_meta=(
                    # Literal string statically defined from a literal string.
                    HintPithSatisfiedMetadata(
                        'All overgrown with azure moss and flowers'),
                    # Literal string dynamically defined from the concatenation
                    # of multiple literal strings, inducing type inference in
                    # PEP 675.
                    HintPithSatisfiedMetadata(
                        'So sweet, ' + 'the sense faints picturing them! ' + 'Thou'
                    ),
                    # Non-literal string dynamically defined from a literal
                    # integer. Ideally, @beartype should reject this
                    # non-literal; pragmatically, runtime type-checkers are
                    # incapable of doing so in the general case. For example,
                    # str() could have been shadowed by a user-defined function
                    # of the same name returning a valid literal string.
                    HintPithSatisfiedMetadata(str(0xFEEDFACE)),

                    # Literal byte string.
                    HintPithUnsatisfiedMetadata(
                        pith=b"For whose path the Atlantic's level powers",
                        # Match that the exception message raised for this
                        # object declares the types *NOT* satisfied by this
                        # object.
                        exception_str_match_regexes=(
                            r'\bstr\b',
                        ),
                        # Match that the exception message raised for this
                        # object does *NOT* contain a newline or bullet
                        # delimiter.
                        exception_str_not_match_regexes=(
                            r'\n',
                            r'\*',
                        ),
                    ),
                ),
            ),
        ))

    # ..................{ RETURN                             }..................
    # Return this list of all PEP-specific type hint metadata.
    return hints_pep_meta



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **type hint reducers** (i.e., low-level callables converting type
hints from one format into another, either losslessly or in a lossy manner).

Type hint reductions imposed by this submodule are purely internal to
:mod:`beartype` itself and thus transient in nature. These reductions are *not*
permanently applied to the ``__annotations__`` dunder dictionaries of the
classes and callables annotated by these type hints.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.meta import URL_ISSUES
from beartype.roar import BeartypeDecorHintRecursionException
from beartype.typing import Optional
from beartype._cave._cavemap import NoneTypeOr
from beartype._check.convert._reduce._redmap import (
    HINT_SIGN_TO_REDUCE_HINT_CACHED_get,
    HINT_SIGN_TO_REDUCE_HINT_UNCACHED_get,
)
from beartype._check.convert._reduce._redrecurse import (
    is_hint_recursive,
    make_hint_sane_recursable,
)
from beartype._check.metadata.metadecor import BeartypeDecorMeta
from beartype._check.metadata.hint.hintsane import (
    HINT_IGNORABLE,
    HINT_SANE_IGNORABLE,
    HintOrSane,
    HintSane,
)
from beartype._conf.confmain import BeartypeConf
from beartype._conf.confcommon import BEARTYPE_CONF_DEFAULT
from beartype._data.hint.sign.datahintsigncls import HintSign
from beartype._data.kind.datakindiota import SENTINEL
from beartype._data.typing.datatypingport import Hint
from beartype._data.typing.datatyping import (
    DictStrToAny,
    HintSignOrNoneOrSentinel,
    TypeStack,
)
from beartype._util.func.arg.utilfuncargiter import ArgKind
from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none
from beartype._util.kind.maplike.utilmapset import remove_mapping_keys

# ....................{ REDUCERS                           }....................
def reduce_hint(
    # Mandatory parameters.
    hint: Hint,

    # Optional keyword-only parameters.
    *,
    arg_kind: Optional[ArgKind] = None,
    cls_stack: TypeStack = None,
    conf: BeartypeConf = BEARTYPE_CONF_DEFAULT,
    decor_meta: Optional[BeartypeDecorMeta] = None,
    hint_parent_sane: Optional[HintSane] = None,
    hint_sign_seed: HintSignOrNoneOrSentinel = SENTINEL,
    is_hint_ignorable_preserved: bool = False,
    pith_name: Optional[str] = None,
    reductions_count: int = 0,
    exception_prefix: str = '',
) -> HintSane:
    '''
    Lower-level type hint reduced (i.e., converted) from the passed higher-level
    type hint if this hint is reducible *or* this hint as is otherwise (i.e., if
    this hint is irreducible).

    This reducer *cannot* be meaningfully memoized, since multiple passed
    parameters (e.g., ``pith_name``, ``cls_stack``) are typically isolated to a
    handful of callables across the codebase currently being decorated by
    :mod:`beartype`. Memoizing this reducer would needlessly consume space and
    time. To improve efficiency, this reducer is instead implemented in terms of
    two lower-level private reducers:

    * The memoized :func:`._reduce_hint_cached` reducer, responsible for
      efficiently reducing *most* (but not all) type hints.
    * The unmemoized :func:`._reduce_hint_uncached` reducer, responsible for
      inefficiently reducing the small subset of type hints contextually
      requiring these problematic parameters.

    Parameters
    ----------
    hint : Hint
        Type hint to be possibly reduced.
    arg_kind : Optional[ArgKind]
        Either:

        * If this hint annotates a parameter of some callable, that parameter's
          **kind** (i.e., :class:`.ArgKind` enumeration member conveying the
          syntactic class of that parameter, constraining how the callable
          declaring that parameter requires that parameter to be passed).
        * Else, :data:`None`.

        Defaults to :data:`None`.
    cls_stack : TypeStack, optional
        **Type stack** (i.e., either a tuple of the one or more
        :func:`beartype.beartype`-decorated classes lexically containing the
        class variable or method annotated by this hint *or* :data:`None`).
        Defaults to :data:`None`.
    conf : BeartypeConf, optional
        **Beartype configuration** (i.e., self-caching dataclass encapsulating
        all settings configuring type-checking for the passed object). Defaults
        to the default beartype configuration.
    decor_meta : Optional[BeartypeDecorMeta], optional
        Either:

        * If this hint annotates a parameter or return of some callable, the
          :mod:`beartype`-specific decorator metadata describing that callable.
        * Else, :data:`None`.

        Defaults to :data:`None`.
    hint_parent_sane : Optional[HintSane], default: None
        Either:

        * If the passed hint is a **root** (i.e., top-most parent hint of a tree
          of child hints), :data:`None`.
        * Else, the passed hint is a **child** of some parent hint. In this
          case, the **sanified parent type hint metadata** (i.e., immutable and
          thus hashable object encapsulating *all* metadata previously returned
          by :mod:`beartype._check.convert.convmain` sanifiers after
          sanitizing the possibly PEP-noncompliant parent hint of this child
          hint into a fully PEP-compliant parent hint).

        Defaults to :data:`None`.
    hint_sign_seed :  HintSignOrNoneOrSentinel, default: SENTINEL
        **Type hint seed sign** (i.e., sign identifying this hint with respect
        to the first reduction performed by this sanification) if this hint is
        ambiguously identifiable by two or more signs *or* the sentinel
        otherwise (i.e., if this hint is uniquely identifiable by one sign).

        This sign is used to seed (i.e., initialize) the first reduction
        internally performed by this sanification, which otherwise defaults to
        the sign returned by the :func:`.get_hint_pep_sign_or_none` getter. This
        parameter should only be passed to handle edge cases in which a hint is
        ambiguously identifiable by two or more signs, including:

        * **Typed dictionary generics** (i.e., user-defined types subclassing
          both the :pep:`484`-compliant :class:`typing.Generic` superclass and
          :pep:`589`-compliant :class:`typing.TypedDict` superclass), which are
          identifiable as both generics *and* typed dictionaries.

        Defaults to the sentinel.
    is_hint_ignorable_preserved : bool, default: False
        Either:

        * If the caller prefers that ignorable hints reduced to a unique
          :data:`.HintSane` object *not* equal to the standard
          :data:`.HINT_SANE_IGNORABLE` singleton but instead encapsulating the
          :data:`.HINT_IGNORABLE` type hint and unique metadata describing the
          ignored hint reduce to that :data:`.HintSane` object rather than the
          :data:`.HINT_SANE_IGNORABLE` singleton, data:`True`.
        * If the caller prefers that ignorable hints reduced to a unique
          :data:`.HintSane` object *not* equal to the standard
          :data:`.HINT_SANE_IGNORABLE` singleton be transparently reduced to the
          :data:`.HINT_SANE_IGNORABLE` singleton, data:`False`. This preference
          is substantially easier for callers to handle but also technically
          lossy, as all unique metadata associated with this reduction is lost.

        Defaults to :data:`False`, as most callers neither require nor desire
        this distinction and are thus incapable of handling ignorable hints
        reduced to unique :data:`.HintSane` objects *not* equal to the standard
        :data:`.HINT_SANE_IGNORABLE` singleton. Most callers only expect the
        :data:`.HINT_SANE_IGNORABLE` singleton.
    pith_name : Optional[str], default: None
        Either:

        * If this hint annotates a parameter of some callable, the name of that
          parameter.
        * If this hint annotates the return of some callable, ``"return"``.
        * Else, :data:`None`.

        Defaults to :data:`None`.
    reductions_count : int, default: None
        Current number of total reductions internally performed by *all* calls
        to this function rooted at this function in the current call stack,
        guarding against accidental infinite recursion between lower-level
        reducers and this higher-level function. Defaults to 0.
    exception_prefix : str, optional
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Returns
    -------
    HintSane
        Either:

        * If this hint is ignorable, :data:`.HINT_SANE_IGNORABLE`.
        * Else if this unignorable hint is reducible to another hint, metadata
          encapsulating this reduction.
        * Else, this unignorable hint is irreducible. In this case, metadata
          encapsulating this hint unmodified.

    Raises
    ------
    BeartypeDecorHintRecursionException
        If the number of total reductions internally performed by the current
        call to this function exceeds the maximum. This exception guards against
        accidental infinite recursion between lower-level hint-specific reducers
        internally called by this higher-level hint-agnostic reducer.
    '''

    # ....................{ PREAMBLE                       }....................
    assert isinstance(arg_kind, NoneTypeOr[ArgKind]), (
        f'{repr(arg_kind)} neither argument kind nor "None".')
    assert isinstance(cls_stack, NoneTypeOr[tuple]), (
        f'{repr(cls_stack)} neither tuple nor "None".')
    assert isinstance(conf, BeartypeConf), f'{repr(conf)} not configuration.'
    assert isinstance(decor_meta, NoneTypeOr[BeartypeDecorMeta]), (
        f'{repr(hint_parent_sane)} neither decoration metadata nor "None".')
    assert isinstance(hint_parent_sane, NoneTypeOr[HintSane]), (
        f'{repr(hint_parent_sane)} neither sanified hint metadata nor "None".')
    assert isinstance(is_hint_ignorable_preserved, bool), (
        f'{repr(is_hint_ignorable_preserved)} not boolean.')
    assert isinstance(pith_name, NoneTypeOr[str]), (
        f'{repr(pith_name)} neither string nor "None".')
    assert isinstance(reductions_count, int), (
        f'{repr(reductions_count)} not integer.')
    assert isinstance(exception_prefix, str), (
        f'{repr(exception_prefix)} not string.')

    # ....................{ LOCALS                         }....................
    # Original unreduced hint passed to this reducer, preserved so as to be
    # embedded in human-readable exception messages.
    hint_old = hint

    # Currently reduced instance of this hint.
    hint_curr: Hint = hint

    # Currently reduced instance of either this hint *OR* metadata encapsulating
    # the sanification of this hint, initialized to this unreduced hint.
    hint_or_sane_curr: HintOrSane = hint

    # Previously reduced instance of either this hint *OR* metadata
    # encapsulating this hint, initialized to this unreduced hint.
    hint_or_sane_prev: HintOrSane = hint

    # Delete the passed "hint" parameter for safety. Permitting this parameter
    # to exist would only promote subtle lexical issues below, where the local
    # variable "hint_curr" is strongly preferred for disambiguity.
    del hint

    # ....................{ SEARCH                         }....................
    # Repeatedly reduce this hint to increasingly irreducible hints until this
    # hint is no longer reducible. This algorithm iteratively reduces this hint
    # with a battery of increasingly non-trivial reductions. For efficiency,
    # reductions are intentionally ordered from most to least efficient.
    while True:
        # ....................{ REDUCE                     }....................
        #FIXME: [SPEED] Optimize into a "while" loop, please. *sigh*
        # For each lower-level reducer...
        for hint_reducer in _HINT_REDUCERS:
            # print(f'[reduce_hint] Reducing {hint_curr} with parent {hint_parent_sane} via {hint_reducer}...')

            # Either:
            # * If this reducer reduces this hint:
            #   * If this reduction produced supplementary metadata, metadata
            #     encapsulating the reduction of this hint by this reducer.
            #   * Else, the reduced hint reduced by this reducer.
            # * Else, this unreduced hint as is.
            hint_or_sane_curr = hint_reducer(
                hint=hint_curr,
                hint_parent_sane=hint_parent_sane,
                arg_kind=arg_kind,
                cls_stack=cls_stack,
                conf=conf,
                decor_meta=decor_meta,
                hint_sign_seed=hint_sign_seed,
                pith_name=pith_name,
                reductions_count=reductions_count,
                exception_prefix=exception_prefix,
            )
            # print(f'[reduce_hint] Reduced to {hint_or_sane_curr}!')

            # If this reduced hint is *NOT* this unreduced hint, this reducer
            # reduced this hint. Halt reducing by these lower-level reducers,
            # enabling the outer loop to decide whether to continue reducing.
            if hint_or_sane_curr is not hint_curr:
                # If this hint reduces to the ignorable "HINT_SANE_IGNORABLE"
                # metadata singleton, then halt reducing immediately.
                #
                # Note that this is merely an optimization avoiding unnecessary
                # iteration. Without this test, hints reduced to this ignorable
                # singleton would require an additional loop through the
                # "_HINT_REDUCERS" tuple. This test elides that iteration.
                if hint_or_sane_curr is HINT_SANE_IGNORABLE:
                    # print(f'[reduce_hint] Ignorably reduced!')
                    return HINT_SANE_IGNORABLE
                # Else, this hint is currently unignorable.
                # print(f'[reduce_hint] Incrementally reduced!')

                # Halt reducing immediately.
                break
            # Else, this unreduced hint remains unmodified. Since this reducer
            # failed to reduce this hint, silently continue to the next reducer.
        # If the above iteration failed to "break", then this unreduced hint
        # remains unmodified across all lower-level reducers. This implies this
        # hint to now be irreducible. Halt reducing immediately.
        else:
            # print(f'[reduce_hint] Irreducible!')
            break
        # Else, the above iteration hit a "break". This hint was reduced by a
        # lower-level reducer above, implying that this hint *COULD* still be
        # reducible. Silently continue reducing.

        # ....................{ RESPOND                    }....................
        # Respond to the lower-level reduction performed above.

        # If reducing this hint generated supplementary metadata...
        if isinstance(hint_or_sane_curr, HintSane):
            # Extract the currently reduced hint from this metadata.
            hint_curr = hint_or_sane_curr.hint

            #FIXME: Should probably be performed down below outside this loop.
            # If this hint reduces to the ignorable "HINT_IGNORABLE"
            # singleton, then halt reducing immediately.
            #
            # Note that this is *NOT* merely an optimization avoiding
            # unnecessary iteration as above. While similar, this logic is
            # distinct from that above. This edge case arises when a reducer
            # avoids reducing to an ignorable hint to the higher-level ignorable
            # "HINT_SANE_IGNORABLE" metadata singleton but instead encapsulates
            # the lower-level "HINT_IGNORABLE" type hint singleton with a new
            # "HintSane" object providing unique metadata describing the ignored
            # type hint. That unique metadata enables parent reducers to
            # selectively decide how to handle ignorable child type hints.
            #
            # Examples include:
            # * The reduce_hint_pep484604() reducer for union type hints.
            if hint_curr is HINT_IGNORABLE:
                # Return either...
                return (
                    # If the caller requests that unique "HintSane" objects
                    # encapsulating the "HINT_IGNORABLE" singleton be preserved,
                    # do so by returning this metadata as is. Note that this is
                    # *NOT* what most callers expect and thus not the default;
                    hint_or_sane_curr
                    if is_hint_ignorable_preserved else
                    # Else, reduce this metadata to the standard ignorable
                    # "HINT_SANE_IGNORABLE" singleton describing ignorable
                    # hints. Note that this is what *MOST* callers expect and
                    # thus the default.
                    HINT_SANE_IGNORABLE
                )
            # Else, this hint does *NOT* reduce to the ignorable
            # "HINT_IGNORABLE" singleton. Ergo, this hint is unignorable.

            # Replace the sanified type hint metadata of the parent hint of this
            # hint by the sanified type hint metadata of this hint itself. Doing
            # so ensures that the next reducer passed the "hint_parent_sane"
            # parameter preserves this metadata during its reduction. Since the
            # most recent reducer call received the prior "hint_parent_sane"
            # parameter, that reducer has already safely preserved the parent
            # metadata by compositing that metadata into this
            # "hint_or_sane_curr" metadata that that reducer returned. Srsly.
            hint_parent_sane = hint_or_sane_curr
        # Else, reducing this hint did *NOT* generate supplementary metadata,
        # implying "hint_or_sane_curr" to be the currently reduced hint. In this
        # case, record this currently reduced hint.
        else:
            hint_curr = hint_or_sane_curr

        #FIXME: Should probably be performed above the prior "if" conditional.
        #FIXME: Currently unused, but useful. Could be required at some point.
        # If this currently reduced hint is exactly the previously reduced hint,
        # the above reducers failed to reduce this hint. Halt reducing entirely.
        #
        # Note that this is a rare (albeit valid) edge case that arises for
        # reducers that unconditionally create and return new... The above
        # "else:" block of the above "for hint_reducer in _HINT_REDUCERS:" loop
        # if hint_or_sane_curr == hint_or_sane_prev:
        #     break

        # ....................{ RECURSION                  }....................
        # Guard against infinite recursion in lower-level reductions with
        # human-readable exceptions.

        # Increment the current number of total reductions internally performed
        # by this call *BEFORE* detecting accidental recursion below.
        reductions_count += 1

        #FIXME: Unit test this, please. No idea how yet. I sigh. *sigh*
        # If the current number of total reductions internally performed
        # by this call exceeds the maximum, raise an exception.
        #
        # Note that this should *NEVER* happen, but probably nonetheless will.
        if reductions_count >= _REDUCTIONS_COUNT_MAX:  # pragma: no cover
            raise BeartypeDecorHintRecursionException(
                f'{exception_prefix}type hint {repr(hint_old)} irreducible. '
                f'Recursion detected when reducing between reduced type hints '
                f'{repr(hint_or_sane_curr)} and {repr(hint_or_sane_prev)}. '
                f'Please submit this exception traceback as a new issue '
                f'to our friendly issue tracker:\n'
                f'\t{URL_ISSUES}\n'
                f'Beartype thanks you for your noble (yet ultimately tragic) '
                f'sacrifice.'
            )
        # Else, the current number of total reductions internally performed
        # by this call is still less than the maximum. In this case, continue.

        # ....................{ PREPARE                    }....................
        # Prepare for the next iterative reduction of this "while" loop.

        # Previously reduced instance of this hint.
        hint_or_sane_prev = hint_or_sane_curr

        # Currently reduced instance of this hint, reverting back to the
        # currently visited hint in preparation for subsequent reduction.
        hint_or_sane_curr = hint_curr

    # ....................{ RETURN                         }....................
    # If this hint is *NOT* already sanified type hint metadata, this hint is
    # unignorable. Why? Because, if this hint were ignorable, this hint would
    # have been reduced to the "HINT_SANE_IGNORABLE" singleton. In this case...
    if not isinstance(hint_or_sane_curr, HintSane):
        # Encapsulate this hint with such metadata, defined as either...
        hint_or_sane_curr = (
            # If this hint has *NO* parent, this is a root hint. In this case,
            # the trivial metadata shallowly encapsulating this root hint;
            HintSane(hint_or_sane_curr)
            if hint_parent_sane is None else
            # Else, this hint has a parent. In this case, the non-trivial
            # metadata deeply encapsulating both this non-root hint *AND* all
            # metadata already associated with this parent hint.
            hint_parent_sane.permute_sane(hint=hint_or_sane_curr)
        )
    # Else, this hint is already sanified type hint metadata. In this case,
    # preserve this metadata as is.

    # Return this possibly reduced hint.
    return hint_or_sane_curr


def reduce_hint_child(hint: Hint, kwargs: DictStrToAny) -> HintSane:
    '''
    Lower-level child type hint reduced (i.e., converted) from the passed
    higher-level child type hint if reducible *or* this child type hint as is
    otherwise (i.e., if this child type hint is irreducible).

    This reducer is a convenience wrapper for the more general-purpose
    :func:`.reduce_hint` reducer, simplifying calls to that reducer when passed
    child hints.

    Parameters
    ----------
    hint : Hint
        Child type hint to be reduced.
    kwargs : DictStrToAny
        Keyword parameters to be passed after being unpacked to the lower-level
        :func:`.reduce_hint` reducer. For safety, this reducer silently ignores
        keyword parameters inapplicable to child hints. This includes:

        * ``arg_kind``, applicable *only* to root hints directly annotating
          callable parameters.
        * ``decor_meta``, applicable *only* to root hints directly annotating
          callable parameters or returns.
        * ``pith_name``, applicable *only* to root hints directly annotating
          callable parameters or returns.

    Returns
    -------
    HintSane
        Either:

        * If this hint is ignorable, :data:`.HINT_SANE_IGNORABLE`.
        * Else if this unignorable hint is reducible to another hint, metadata
          encapsulating this reduction.
        * Else, this unignorable hint is irreducible. In this case, metadata
          encapsulating this hint unmodified.
    '''

    # Remove all unsafe keyword parameters (i.e., parameters that are
    # inapplicable to child hints and thus *NOT* safely passable to the
    # subsequently called reduce_hint() function) from this dictionary.
    remove_mapping_keys(kwargs, _REDUCE_HINT_CHILD_ARG_NAMES_UNSAFE)

    # Return this child hint possibly reduced to a lower-level hint.
    return reduce_hint(hint=hint, **kwargs)

# ....................{ PRIVATE ~ reducers                 }....................
def _reduce_hint_cached(
    hint: Hint,
    hint_sign_seed: HintSignOrNoneOrSentinel,
    exception_prefix: str,
    **kwargs
) -> HintOrSane:
    '''
    Lower-level type hint reduced (i.e., converted) from the passed higher-level
    type hint if this hint is reducible by a **memoized reducer** (i.e.,
    lower-level reducer accepting *only* a passed hint and thus readily amenable
    to memoization) *or* this hint as is otherwise (i.e., if this hint is *not*
    reducible by a memoized reducer).

    Parameters
    ----------
    hint : Hint
        Type hint to be possibly reduced.
    hint_sign_seed : HintSignOrNoneOrSentinel
        Sign with which to seed (i.e., initialize) this reduction. See also the
        :func:`.reduce_hint` docstring for further details.
    exception_prefix : str
        Human-readable substring prefixing raised exception messages.

    All remaining keyword parameters are silently ignored.

    Returns
    -------
    HintOrSane
        Either:

        * If this hint is ignorable, :data:`.HINT_SANE_IGNORABLE`.
        * Else if this unignorable hint is reducible to another hint by a
          memoized reducer, metadata encapsulating this reduction.
        * Else, this hint unmodified as is.
    '''
    assert (
        hint_sign_seed is SENTINEL or
        isinstance(hint_sign_seed, NoneTypeOr[HintSign])
    ), (f'{repr(hint_sign_seed)} neither hint sign, "None", nor sentinel.')

    # Reduced hint to be returned, defaulting to the passed unreduced hint.
    hint_or_sane: HintOrSane = hint

    # Sign uniquely identifying this hint if this hint is PEP-compliant *OR*
    # "None" otherwise (e.g., if this hint is PEP-noncompliant), defined as
    # either...
    hint_sign = (
        # If the caller did *NOT* explicitly pass a sign with which to seed this
        # reduction, the standard sign uniquely identifying this hint;
        get_hint_pep_sign_or_none(hint)
        if hint_sign_seed is SENTINEL else
        # Else, the caller explicitly passed a sign with which to seed this
        # reduction. In this case, that sign.
        hint_sign_seed
    )

    # Memoized reducer reducing this hint if any *OR* "None" otherwise.
    hint_reducer_cached = HINT_SIGN_TO_REDUCE_HINT_CACHED_get(hint_sign)  # type: ignore[arg-type]

    # If a memoized reducer reduces this hint...
    if hint_reducer_cached is not None:
        # print(f'[_reduce_hint_cached] Reducing cached hint {repr(hint)}...')

        #FIXME: [SPEED] Is there any point to passing the "exception_prefix"
        #parameter? Possibly. Not sure. Isn't this parameter a constant? No?
        #Does it actually vary with context? Can't recall. Investigate up!

        # Reduce this hint by calling this reducer.
        #
        # Note that parameters are intentionally passed positionally to this
        # possibly memoized callable prohibiting keyword parameters.
        hint_or_sane = hint_reducer_cached(hint, exception_prefix)
    # Else, *NO* memoized reducer reduces this hint. In this case, preserve this
    # hint as is.

    # Return this possibly reduced hint.
    return hint_or_sane


def _reduce_hint_uncached(
    hint: Hint,
    hint_sign_seed: HintSignOrNoneOrSentinel,
    **kwargs
) -> HintOrSane:
    '''
    Lower-level type hint reduced (i.e., converted) from the passed higher-level
    type hint if this hint is reducible by a **unmemoized reducer** (i.e.,
    lower-level reducer accepting *only* a passed hint and thus readily amenable
    to memoization) *or* this hint as is otherwise (i.e., if this hint is *not*
    reducible by a unmemoized reducer).

    Parameters
    ----------
    hint : Hint
        Type hint to be possibly reduced.
    hint_sign_seed : HintSignOrNoneOrSentinel
        Sign with which to seed (i.e., initialize) this reduction. See also the
        :func:`.reduce_hint` docstring for further details.

    All remaining keyword parameters are silently ignored.

    Returns
    -------
    HintOrSane
        Either:

        * If this hint is ignorable, :data:`.HINT_SANE_IGNORABLE`.
        * Else if this unignorable hint is reducible to another hint by a
          unmemoized reducer, metadata encapsulating this reduction.
        * Else, this hint unmodified as is.
    '''
    assert (
        hint_sign_seed is SENTINEL or
        isinstance(hint_sign_seed, NoneTypeOr[HintSign])
    ), (f'{repr(hint_sign_seed)} neither hint sign, "None", nor sentinel.')

    # Reduced hint to be returned, defaulting to the passed unreduced hint.
    hint_or_sane: HintOrSane = hint

    # Sign uniquely identifying this hint if this hint is PEP-compliant *OR*
    # "None" otherwise (e.g., if this hint is PEP-noncompliant), defined as
    # either...
    hint_sign = (
        # If the caller did *NOT* explicitly pass a sign with which to seed this
        # reduction, the standard sign uniquely identifying this hint;
        get_hint_pep_sign_or_none(hint)
        if hint_sign_seed is SENTINEL else
        # Else, the caller explicitly passed a sign with which to seed this
        # reduction. In this case, that sign.
        hint_sign_seed
    )

    # Unmemoized reducer reducing this hint if any *OR* "None" otherwise.
    hint_reducer_uncached = HINT_SIGN_TO_REDUCE_HINT_UNCACHED_get(hint_sign)  # type: ignore[arg-type]

    # If a unmemoized reducer reduces this hint...
    if hint_reducer_uncached is not None:
        # print(f'[_reduce_hint_cached] Reducing cached hint {repr(hint)}...')

        # Reduce this hint by calling this reducer.
        hint_or_sane = hint_reducer_uncached(hint=hint, **kwargs)
    # Else, *NO* unmemoized reducer reduces this hint. In this case, preserve
    # this hint as is.

    # Return this possibly reduced hint.
    return hint_or_sane


def _reduce_hint_overrides(
    hint: Hint,
    conf: BeartypeConf,
    hint_parent_sane: Optional[HintSane],
    **kwargs
) -> HintOrSane:
    '''
    Lower-level type hint reduced (i.e., converted) from the passed higher-level
    type hint if this hint is reducible as a **hint override** (i.e., key of
    the :attr:`.BeartypeConf.hint_overrides` dictionary of the passed beartype
    configuration) *or* this hint as is otherwise (i.e., if this hint is *not*
    as a hint override).

    Parameters
    ----------
    hint : Hint
        Type hint to be possibly reduced.
    conf : BeartypeConf
        **Beartype configuration** (i.e., self-caching dataclass encapsulating
        all settings configuring type-checking for the passed object).
    hint_parent_sane : Optional[HintSane]
        Either:

        * If the passed hint is a **root** (i.e., top-most parent hint of a tree
          of child hints), :data:`None`.
        * Else, the passed hint is a **child** of some parent hint. In this
          case, the **sanified parent type hint metadata** (i.e., immutable and
          thus hashable object encapsulating *all* metadata previously returned
          by :mod:`beartype._check.convert.convmain` sanifiers after
          sanitizing the possibly PEP-noncompliant parent hint of this child
          hint into a fully PEP-compliant parent hint).

    All remaining keyword parameters are silently ignored.

    Returns
    -------
    HintOrSane
        Either:

        * If this hint has already been overridden by a prior recursive
          reduction in the current tree of reductions, this hint before being
          overridden yet again (i.e., this hint as is).
        * Else if this unignorable hint is overridden by another hint, metadata
          encapsulating this override.
        * Else, this unignorable hint is *not* overridden by another hint, this
          hint as is.
    '''

    # Overridden hint to be returned, defaulting to the passed un-overridden
    # hint for safety and simplicity.
    hint_or_sane: HintOrSane = hint

    # Attempt to...
    #
    # Note that the is_object_hashable() tester is internally implemented with
    # the same Easier to Ask for Permission than Forgiveness (EAFP)-based
    # "try-except" block and is thus equally inefficient. In fact, the current
    # approach avoids an extraneous call to that tester and is thus marginally
    # faster. (Emphasis on "marginally.")
    try:
        # Hint overriding this hint if this configuration overrides this hint
        # *OR* the sentinel otherwise (i.e., if this hist is *NOT* overridden).
        #
        # Note that this raises "TypeError" when this hint is unhashable.
        # print(f'Overriding hint {repr(hint)} via {repr(conf.hint_overrides)}...')
        hint_overridden = conf.hint_overrides.get(hint, SENTINEL)

        # If neither...
        if not (
            # This hint is not overridden *NOR*...
            hint_overridden is SENTINEL or
            # Else, this hint is overridden.
            #
            # If this overridden hint is recursive, this hint has already been
            # overridden by a previously performed reduction. Avoid attempting
            # to reoverride this hint again with the same hint override; doing
            # so would provoke infinite recursion. Instead, preserve this
            # un-overridden hint by returning this hint as is.
            #
            # Certainly, various approaches to type-checking recursive hints
            # exists. @beartype currently embraces the easiest, fastest, and
            # laziest approach: just ignore all recursion! \o/
            #
            # Note that:
            # * This tester raises "TypeError" when this hint is unhashable.
            # * This tester intentionally accepts the default value "0" for the
            #   optional parameter "hint_recursable_depth_max", ensuring this
            #   overridden hint is considered to be recursive when this
            #   overridden hint has already been overridden a single time.
            #   Unlike comparable kinds of recursable hints (e.g., PEP
            #   695-compliant type aliases), hint overrides typically convey
            #   *NO* internal structure and thus merit *NO* deeper recursion.
            #   Hint overrides instruct @beartype to perform simple
            #   global-search-and-replacements on exactly matching type hints.
            #   Recursion is neither desirable nor necessary.
            #
            #   Consider the prototypical hint overrides of
            #   "BeartypeConf(hint_overrides={float: float | int})". After
            #   expanding the builtin "float" type to the PEP 604-compliant
            #   union "float | int", attempting to recursively re-apply the same
            #   expansion silently reduces to a noop (e.g., "float | int"
            #   expands to "float | float | int", equal to "float | int").
            is_hint_recursive(hint=hint, hint_parent_sane=hint_parent_sane)
        ):
            # Then this overridden hint is *NOT* recursive, implying this hint
            # *CANNOT* have already been overridden by a previously performed
            # reduction. Why? Because an overridden hint revisited by the
            # current breadth-first search would by definition by recursive.

            # Metadata guarding this hint against infinite recursion, recording
            # this hint as already having been overridden *BEFORE* reducing and
            # thus forgetting this hint.
            #
            # Note that this intentionally replaces the metadata encapsulating
            # the sanification of the parent hint of this hint by the metadata
            # encapsulating the sanification of this hint itself. Doing so
            # ensures that the next reducer passed the "hint_parent_sane"
            # parameter preserves this metadata during its reduction. Since the
            # most recent reducer call received the prior "hint_parent_sane"
            # parameter, that reducer has already safely preserved the parent
            # metadata by compositing that metadata into this
            # "hint_or_sane_curr" metadata that that reducer returned. Srsly.
            hint_or_sane = make_hint_sane_recursable(
                # The recursable form of this overridden hint is the
                # pre-overridden hint tested above by the is_hint_recursive()
                # recursion guard.
                hint_recursable=hint,
                # The non-recursable form of this overridden hint is the
                # overridden hint encapsulated by the metadata returned by this
                # factory.
                hint_nonrecursable=hint_overridden,
                hint_parent_sane=hint_parent_sane,
            )
        # Else, this overridden hint is recursive. In this case, preserve this
        # un-overridden hint rather than reducing this hint to the ignorable
        # "HINT_SANE_IGNORABLE" singleton. Why? Because un-overridden hints are
        # themselves valid type hints and thus have semantic meaning in and of
        # themselves (e.g., the "float" in the hint override
        # "BeartypeConf(hint_overrides={float: float | int})" has semantic
        # meaning as a builtin type).
    # If doing so raises a "TypeError", this hint is unhashable and thus
    # inapplicable for hint overriding. In this case, preserve this hint as is.
    except TypeError:
        pass

    # Return this possibly overridden hint.
    return hint_or_sane

# ....................{ PRIVATE ~ globals                  }....................
_HINT_REDUCERS = (
    # ....................{ PHASE ~ override               }....................
    # Attempt to reduce this hint to another hint configured by a user-defined
    # hint override *BEFORE* applying standard reductions. User preference
    # assumes precedence over standard precedent. Nice one-liner, huh? </heh>
    _reduce_hint_overrides,

    # ....................{ PHASE ~ context-free           }....................
    # Attempt to reduce this hint with a context-free reduction *BEFORE*
    # reducing this hint with a contextual reduction. Due to *NOT* depending on
    # contextual state, context-free reductions are readily memoizable and thus
    # faster than contextual reductions.
    _reduce_hint_cached,

    # ....................{ PHASE ~ contextual             }....................
    # Attempt to reduce this hint with a contextual reduction. Due to depending
    # on contextual state, contextual reductions are *NOT* be readily memoizable
    # and thus slower than context-free reductions.
    _reduce_hint_uncached,
)
'''
Tuple of all private high-level reducers defined by this submodule above.
'''


_REDUCTIONS_COUNT_MAX = 64
'''
Maximum number of total reductions internally performed by each call of the
:func:`reduce_hint` function.

This constant is a relatively arbitrary magic number selected so as to guard
against accidental infinite recursion between lower-level PEP-specific reducers
internally called by :func:`reduce_hint`.
'''


_REDUCE_HINT_CHILD_ARG_NAMES_UNSAFE = frozenset((
    # Applicable *ONLY* to root hints directly annotating callable parameters.
    'arg_kind',

    # Applicable *ONLY* to root hints directly annotating callable parameters
    # or returns.
    'decor_meta',
    'pith_name',
))
'''
Frozen set of the names of all **unsafe child type hint reducer keyword
parameters** (i.e., keyword parameters inapplicable to child type hints and thus
*not* safely passable from the higher-level :func:`.reduce_hint_child` to
lower-level :func:`.reduce_hint` reducer.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype decorator :pep:`577` unit tests.

This submodule unit tests :pep:`577` support implemented in the
:func:`beartype.beartype` decorator.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_decor_pep557() -> None:
    '''
    Test :pep:`557` support implemented in the :func:`beartype.beartype`
    decorator if the active Python interpreter targets Python >= 3.8 *or* skip
    otherwise.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer test-specific imports.
    from beartype import (
        BeartypeConf,
        beartype,
    )
    from beartype.roar import (
        BeartypeCallHintPep557FieldViolation,
        BeartypeCallHintParamViolation,
    )
    from beartype.typing import (
        ClassVar,
        Optional,
    )
    from dataclasses import (
        FrozenInstanceError,
        InitVar,
        dataclass,
        field,
    )
    from pytest import raises

    # ..................{ LOCALS                             }..................
    # Beartype decorator type-checking *ALL* PEP 557-compliant dataclass fields.
    beartype_pep557 = beartype(conf=BeartypeConf(is_pep557_fields=True))

    # Modifiable list of "**kwargs" dictionaries to be passed as the keyword
    # parameters configuring each decoration of the dataclass below by the
    # @dataclass decorator.
    dataclasses_kwargs = [
        # The default parameter-less @dataclass decorator.
        {},

        # The frozen @dataclass decorator.
        dict(frozen=True),

        # The slotted @dataclass decorator, only supported under Python >= 3.10.
        dict(slots=True),
    ]

    # ..................{ DATACLASSES                        }..................
    # For each dictionary of keyword parameters configuring this dataclass...
    for dataclass_kwargs in dataclasses_kwargs:
        @beartype_pep557
        @dataclass(**dataclass_kwargs)
        class SoSolemnSoSerene(object):
            '''
            Arbitrary dataclass type-checked by :func:`beartype.beartype`.
            '''

            # ..................{ PARAMETERS ~ mandatory     }..................
            but_for_such_faith: InitVar[str]
            '''
            Mandatory non-field parameter passed by the :func:`dataclass`
            decorator to both this :meth:`__init__` method *and* the explicit
            :meth:`__post_init__` method defined below.

            Note this mandatory parameter *must* be declared before all optional
            parameters due to inscrutable issues in the :func:`.dataclass`
            decorator. Notably, :func:`.dataclass` refuses to correctly reorder
            parameters. *sigh*
            '''

            # ..................{ PARAMETERS ~ optional      }..................
            that_man_may_be: Optional[str] = None
            '''
            Optional field passed by the :func:`.dataclass` decorator to
            the implicit :meth:`__init__` method synthesized for this dataclass.
            '''


            thou_hast_a_voice: int = field(default=0xBABE, init=False)
            '''
            Optional uninitialized field *not* passed by the :func:`.dataclass`
            decorator to the implicit :meth:`__init__` method synthesized for
            this dataclass.
            '''


            with_nature_reconciled: ClassVar[bool] = True
            '''
            Optional class variable ignored by the :func:`.dataclass` decorator
            and thus *not* passed to the implicit :meth:`__init__` method
            synthesized for this dataclass.
            '''

            # ..................{ INITIALIZERS               }..................
            @beartype
            def __post_init__(self, but_for_such_faith: str) -> None:
                '''
                :func:`dataclasses.dataclass`-specific dunder method implicitly
                called by the :meth:`__init__` method synthesized for this
                class, explicitly type-checked by :func:`beartype.beartype` for
                testing.
                '''

                # If the caller failed to pass this optional field to the
                # __init__() constructor, default this field to this mandatory
                # field required to be passed by the caller.
                #
                # Note that this field *CANNOT* be safely assigned with a simple
                # assignment statement: e.g.,
                #     self.that_man_may_be = but_for_such_faith
                #
                # Doing so suffices for the general case but fails for frozen
                # dataclasses, which enforce immutability even within this
                # __post_init__() initialization method, which arguably
                # constitutes a long-standing bug that CPython devs will
                # probably *NEVER* resolve: e.g.,
                #     dataclasses.FrozenInstanceError: cannot assign to field
                #     'that_man_may_be'
                #
                # Instead, we assign this field by circumventing the
                # __setattr__() dunder method defined for frozen dataclasses
                # entirely in favour of a permissive alternative guaranteed to
                # exist: the object.__setattr__() dunder method. We sigh.
                if self.that_man_may_be is None:
                    object.__setattr__(
                        self, 'that_man_may_be', but_for_such_faith)
                # Else, the caller passed this optional field to the
                # __init__() constructor. In this case, preserve this field.

        # ..................{ LOCALS                         }..................
        # Arbitrary instance of this dataclass exercising all edge cases.
        great_mountain = SoSolemnSoSerene(
            but_for_such_faith='So solemn, so serene, that man may be,')

        # True only if this dataclass is frozen, defaulting to false.
        IS_DATACLASS_FROZEN = dataclass_kwargs.get('frozen', False)

        # True only if this dataclass is slotted, defaulting to false.
        IS_DATACLASS_SLOTTED = dataclass_kwargs.get('slots', False)

        # ..................{ ASSERTS ~ fields               }..................
        # Assert that this dataclass initializes these fields to have the
        # expected initial values.
        assert great_mountain.that_man_may_be == (
            'So solemn, so serene, that man may be,')
        assert great_mountain.thou_hast_a_voice == 0xBABE

        # Assert that attempting to set these fields to new values violating the
        # hints annotating these fields raises the expected exception.
        with raises(BeartypeCallHintPep557FieldViolation):
            great_mountain.that_man_may_be = (
                b'Somewhere between the throne, and where I sit')
        with raises(BeartypeCallHintPep557FieldViolation):
            great_mountain.thou_hast_a_voice = (
                'Here on this spot of earth. Search, Thea, search!')

        # If this dataclass is frozen...
        if IS_DATACLASS_FROZEN:
            # Assert that this dataclass prohibits callers from modifying *ANY*
            # attributes (including both instance and class variables) by
            # raising the expected exception on attempting to do so.
            with raises(FrozenInstanceError):
                great_mountain.that_man_may_be = (
                    'With backward footing through the shade a space:')
            with raises(FrozenInstanceError):
                great_mountain.thou_hast_a_voice = len(
                    "He follow'd, and she turn'd to lead the way")
        # Else, this dataclass is *NOT* frozen. In this case...
        else:
            # Assert that this dataclass allows callers to set these fields to
            # new values satisfying the hints annotating these fields.
            great_mountain.that_man_may_be = (
                'Through aged boughs, that yielded like the mist')
            assert great_mountain.that_man_may_be == (
                'Through aged boughs, that yielded like the mist')
            great_mountain.thou_hast_a_voice = len(
                'Which eagles cleave upmounting from their nest.')
            assert great_mountain.thou_hast_a_voice == len(
                'Which eagles cleave upmounting from their nest.')

        # ..................{ ASSERTS ~ InitVar              }..................
        # Assert that this dataclass does *NOT* expose initialization-only
        # parameters as fields.
        assert not hasattr(great_mountain, 'but_for_such_faith')

        # Assert that attempting to instantiate an instance of this dataclass
        # with a parameter violating the corresponding hint annotating the
        # initialization-only parameter of the same name raises the expected
        # exception.
        with raises(BeartypeCallHintParamViolation):
            SoSolemnSoSerene(but_for_such_faith=0xBEEFBABE)

        # ..................{ ASSERTS ~ ClassVar             }..................
        # Assert that this dataclass initializes these class variables to have
        # the expected initial values.
        assert great_mountain.with_nature_reconciled == True

        # If this dataclass is frozen...
        if IS_DATACLASS_FROZEN:
            # Assert that this dataclass prohibits callers from modifying *ANY*
            # attributes (including both instance and class variables) by
            # raising the expected exception on doing so.
            with raises(FrozenInstanceError):
                great_mountain.with_nature_reconciled = (
                    'Thus brief; then with beseeching eyes she went')
        # Else, this dataclass is *NOT* frozen.
        #
        # If this dataclass is slotted...
        elif IS_DATACLASS_SLOTTED:
            # Assert that this dataclass prohibits callers from modifying *ONLY*
            # class attributes by raising the expected exception on doing so.
            with raises(AttributeError):
                great_mountain.with_nature_reconciled = (
                    'Meanwhile in other realms big tears were shed,')
        # In this case...
        else:
            # Assert that this dataclass permissively allows callers to
            # erroneously modify these class variables to have invalid new
            # values violating the hints annotating these variables. Although
            # non-ideal, type-checking class variables would require @beartype
            # to dangerously monkey-patch the metaclass of dataclasses -- a
            # bridge too far over troubled waters.
            great_mountain.with_nature_reconciled = (
                'My strong identity, my real self,')
            assert great_mountain.with_nature_reconciled == (
                'My strong identity, my real self,')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **Decidedly Object-Oriented Runtime-checking (DOOR) any type hint
classes** (i.e., :class:`beartype.door.TypeHint` subclasses implementing support
for the :pep:`484`-compliant :obj:`typing.Any` singleton type hint).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.door._cls.doorsuper import TypeHint
from beartype.typing import (
    TYPE_CHECKING,
    Any,
)

# ....................{ SUBCLASSES                         }....................
class AnyTypeHint(TypeHint):
    '''
    **Any type hint wrapper** (i.e., high-level object encapsulating the
    low-level :pep:`484`-compliant :obj:`typing.Any` singleton type hint).
    '''

    # ..................{ STATIC                             }..................
    # Squelch false negatives from static type checkers.
    if TYPE_CHECKING:
        _hint: type

    # ..................{ PRIVATE ~ properties               }..................
    @property
    def _is_args_ignorable(self) -> bool:

        # Unconditionally return true, as "typing.Any" is *ALWAYS* unsubscripted
        # and could thus be said to only have ignorable arguments. Semantics.
        return True

    # ..................{ PRIVATE ~ methods                  }..................
    def _is_subhint_branch(self, branch: TypeHint) -> bool:
        # print(f'[AnyTypeHint._is_subhint_branch] Comparing {self} to {branch}...')

        # Unconditionally return false, as "typing.Any" is a subhint of *NO*
        # hint other than itself. However, the following superclass methods
        # already universally handle this common edge case in which the passed
        # hint is "typing.Any":
        # * The public is_subhint() method.
        # * The private _is_subhint_branch() method.
        #
        # The passed hint is thus guaranteed to *NOT* also be "typing.Any", so
        # this hint *CANNOT* be a subhint of that hint.
        return False



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **abstract syntax tree (AST) transformers** (i.e., low-level classes
instrumenting well-typed third-party modules with runtime type-checking
dynamically generated by the :func:`beartype.beartype` decorator).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from ast import (
    ClassDef,
    NodeTransformer,
)
from beartype.claw._ast._scope.clawastscope import BeartypeNodeScope
from beartype.claw._ast._scope.clawastscopes import BeartypeNodeScopes
from beartype.claw._ast._kind.clawastassign import (
    BeartypeNodeTransformerAssignMixin)
from beartype.claw._ast._kind.clawastimport import (
    BeartypeNodeTransformerImportMixin)
from beartype.claw._ast._kind.clawastmodule import (
    BeartypeNodeTransformerModuleMixin)
from beartype.claw._ast._pep.clawastpep695 import (
    BeartypeNodeTransformerPep695Mixin)
from beartype.claw._ast._clawastutil import BeartypeNodeTransformerUtilityMixin
from beartype.roar._roarexc import _BeartypeClawAstNodeScopesException
from beartype.typing import (
    Optional,
)
from beartype._conf.confmain import BeartypeConf
from beartype._data.api.standard.dataast import TYPES_NODE_LEXICAL_SCOPE
from beartype._data.typing.datatyping import (
    ListStrs,
    NodeCallable,
    NodeT,
)
from beartype._util.ast.utilasttest import is_node_callable_typed
from beartype._util.cache.utilcachecall import property_cached

# ....................{ SUBCLASSES                         }....................
#FIXME: Unit test us up, please.
class BeartypeNodeTransformer(
    # PEP-agnostic superclass defining "core" AST node transformation logic.
    NodeTransformer,

    # PEP-agnostic mixins defining supplementary AST node functionality in a
    # PEP-agnostic manner.
    BeartypeNodeTransformerAssignMixin,
    BeartypeNodeTransformerImportMixin,
    BeartypeNodeTransformerModuleMixin,
    BeartypeNodeTransformerUtilityMixin,

    # PEP-specific mixins defining additional AST node transformations in a
    # PEP-specific manner.
    BeartypeNodeTransformerPep695Mixin,
):
    '''
    **Beartype abstract syntax tree (AST) node transformer** (i.e., visitor
    pattern recursively transforming the AST tree passed to the :meth:`visit`
    method by decorating all typed callables and classes by the
    :func:`beartype.beartype` decorator).

    Design
    ------
    This class was largely designed by reverse-engineering the standard
    :mod:`ast` module using the following code snippet. When run as the body of
    a script from the command line (e.g., ``python3 {muh_script}.py``), this
    snippet pretty-prints the desired target AST subtree implementing the
    desired source code (specified in this snippet via the ``CODE`` global). In
    short, this snippet trivializes the definition of arbitrarily complex
    AST-based code from arbitrarily complex Python code:

    .. code-block:: python

       import ast

       # Arbitrary desired code to pretty-print the AST representation of.
       CODE = """
       from beartype import beartype
       from beartype._conf.confcache import beartype_conf_id_to_conf

       @beartype(conf=beartype_conf_id_to_conf[139870142111616])
       def muh_func(): pass
       """

       # Dismantled, this is:
       # * "indent=...", producing pretty-printed (i.e., indented) output.
       # * "include_attributes=True", enabling pseudo-nodes (i.e., nodes lacking
       #   associated code metadata) to be distinguished from standard nodes
       #   (i.e., nodes having such metadata).
       print(ast.dump(ast.parse(CODE), indent=4, include_attributes=True))

    Attributes
    ----------
    _conf : BeartypeConf
        **Beartype configuration** (i.e., dataclass configuring the
        :mod:`beartype.beartype` decorator for *all* decoratable objects
        recursively decorated by this node transformer).
    _module_name : str
        Fully-qualified name of the current module being transformed.
    _scopes : BeartypeNodeScopes
        **Lexical scope stack** (i.e., list of the one or more dataclasses
        aggregating all metadata required to detect and manage lexical scopes
        being recursively visited by this AST transformer).

    See Also
    --------
    https://github.com/agronholm/typeguard/blob/fe5b578595e00d5e45c3795d451dcd7935743809/src/typeguard/importhook.py
        Last commit of the third-party Python package whose
        ``@typeguard.importhook.TypeguardTransformer`` class implements import
        hooks performing runtime type-checking in a similar manner, strongly
        inspiring this implementation.

        Note that all subsequent commits to that package generalize those import
        hooks into something else entirely, which increasingly resembles a
        static type-checker run at runtime; while fascinating and almost
        certainly ingenious, those commits are sufficiently inscrutable,
        undocumented, and unintelligible to warrant caution. Nonetheless, thanks
        so much to @agronholm (Alex Grönholm) for his pulse-pounding innovations
        in this burgeoning field! Our AST transformer is for you, @agronholm.
    '''

    # ..................{ CLASS VARIABLES                    }..................
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: Subclasses declaring uniquely subclass-specific instance
    # variables *MUST* additionally slot those variables. Subclasses violating
    # this constraint will be usable but unslotted, which defeats the purpose.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # Slot all instance variables defined on this object to reduce the costs of
    # both reading and writing these variables by approximately ~10%.
    __slots__ = (
        '_conf',
        '_module_name',
        '_scopes',
    )

    # ..................{ INITIALIZERS                       }..................
    def __init__(
        self,

        # Mandatory keyword-only parameters.
        *,
        module_name: str,
        conf: BeartypeConf,
    ) -> None:
        '''
        Initialize this node transformer.

        Parameters
        ----------
        module_name : str
            Fully-qualified name of the external third-party module being
            transformed by this node transformer.
        conf : BeartypeConf
            **Beartype configuration** (i.e., dataclass configuring the
            :mod:`beartype.beartype` decorator for *all* decoratable objects
            recursively decorated by this node transformer).
        '''
        assert isinstance(module_name, str), (
            f'{repr(module_name)} not string.')
        assert isinstance(conf, BeartypeConf), (
            f'{repr(conf)} not beartype configuration.')

        # Initialize our superclass.
        # print(f'Initializing module "{module_name}" AST transformation...')
        super().__init__()

        # Classify all passed parameters.
        self._conf = conf
        self._module_name = module_name

        # Lexical scope stack, initially containing *ONLY* the default global
        # scope for statements in the body of the current module.
        self._scopes = BeartypeNodeScopes(module_name=module_name)

    # ..................{ PROPERTIES                         }..................
    #FIXME: Unit test us up, please. *sigh*
    @property  # type: ignore[misc]
    @property_cached
    def _module_basenames(self) -> ListStrs:
        '''
        List of the one or more unqualified basenames comprising the
        fully-qualified ``"."``-delimited name of the currently visited module.

        This property is memoized for efficiency.

        Returns
        -------
        list[str]
            List of the one or more unqualified basenames comprising the
            fully-qualified name of the currently visited module.
        '''

        # List of each unqualified basename comprising the name of the currently
        # visited module, split from the fully-qualified name of that module on
        # "." delimiters.
        #
        # Note that the "str.split('.')" and "str.rsplit('.')" calls produce the
        # same lists under all edge cases. We arbitrarily call the former rather
        # than the latter for simplicity.
        module_basenames = self._module_name.split('.')

        # Return this list.
        return module_basenames


    #FIXME: Unit test us up, please. *sigh*
    @property
    def _scope(self) -> BeartypeNodeScope:  # type: ignore[override]
        '''
        **Currently visited lexical scope** (i.e., dataclass aggregating all
        metadata required to detect and manage the lexical scope of the
        currently visited node of the currently visited module) if this
        abstract syntax tree (AST) transformer has already visited at least that
        module's root :class:`ast.Module` node via the :meth:`.visitModule`
        method *or* raise an exception otherwise (i.e., if this property is
        unsafely accessed at an early time).

        Returns
        -------
        BeartypeNodeScope
            :data:`True` only if the current lexical scope is a module scope.

        Raises
        ------
        _BeartypeClawAstNodeScopesException
            If this property is unsafely accessed at an early time.
        '''

        # If the lexical scope stack is empty, this AST transformer has yet to
        # visit the root "Module" node of the currently visited module. In this
        # case, raise an exception.
        if not self._scopes:
            raise _BeartypeClawAstNodeScopesException(
                f'Module "{self._module_name}" '
                f'global lexical scope has yet to be visited '
                f'(i.e., visitModule() method not previously called).'
            )
        # Else, the lexical scope stack is non-empty.

        # Return the last lexical scope on this stack -- the top-most item
        # signifying the currently visited lexical scope.
        return self._scopes[-1]

    # ..................{ SUPERCLASS                         }..................
    # Overridden methods first defined by the "NodeTransformer" superclass.
    def generic_visit(self, node: NodeT) -> NodeT:
        '''
        Recursively visit and possibly transform *all* child nodes of the passed
        parent node in-place (i.e., preserving this parent node as is).

        Parameters
        ----------
        node : NodeT
            Parent node to transform *all* child nodes of.

        Returns
        -------
        NodeT
            Parent node returned and thus preserved as is.
        '''

        # Type of this parent node.
        node_type = type(node)

        # If this parent node declares a new lexical scope (i.e., by defining a
        # new class or callable)...
        if node_type in TYPES_NODE_LEXICAL_SCOPE:
            # Add the type of this parent node to the top of the stack of all
            # current lexical scopes *BEFORE* visiting any child nodes of this
            # parent node.
            self._scopes.append_scope_nested(
                # Fully-qualified name of the parent scope (i.e.,
                # "{self._scopes[-1].name}") followed by the unqualified
                # basename of this new class or callable declaring this new
                # lexical scope (i.e., "{node.name}").
                #
                # Note that both the "ast.ClassDef" *AND* "ast.FunctionDef" node
                # types define the "name" instance variable accessed here.
                name=f'{self._scopes[-1].name}.{node.name}',  # type: ignore[attr-defined]
                node_type=node_type,
            )

            # Recursively visit *ALL* child nodes of this parent node.
            super().generic_visit(node)

            # Remove the type of this parent node from the top of the stack of
            # all current lexical scopes *AFTER* visiting all child nodes of
            # this parent node.
            self._scopes.pop()
        # Else, this parent node does *NOT* declare a new lexical scope. In this
        # case...
        else:
            # Recursively visit all child nodes of this parent node *WITHOUT*.
            # modifying the stack of all current lexical scopes.
            super().generic_visit(node)

        # Return this parent node as is.
        return node

    # ..................{ VISITORS ~ class                   }..................
    #FIXME: Implement us up, please.
    def visit_ClassDef(self, node: ClassDef) -> Optional[ClassDef]:
        '''
        Add a new child node to the passed **class node** (i.e., node
        encapsulating the definition of a pure-Python class) unconditionally
        decorating that class by our private
        :func:`beartype._decor.decorcore.beartype_object_nonfatal` decorator.

        Parameters
        ----------
        node : ClassDef
            Class node to be transformed.

        Returns
        -------
        Optional[ClassDef]
            This same class node.
        '''

        # Add a new child decoration node to this parent class node decorating
        # this class by @beartype under this configuration.
        self._decorate_node_beartype(node=node, conf=self._conf)

        # Recursively transform *ALL* child nodes of this parent class node.
        # Note that doing so implicitly calls the visit_FunctionDef() method
        # (defined below), each of which then effectively reduces to a noop.
        return self.generic_visit(node)

    # ..................{ VISITORS ~ callable                }..................
    def visit_FunctionDef(self, node: NodeCallable) -> Optional[NodeCallable]:
        '''
        Add a new child node to the passed **callable node** (i.e., node
        encapsulating the definition of a pure-Python function or method)
        decorating that callable by our private
        :func:`beartype._decor.decorcore.beartype_object_nonfatal` decorator if
        that callable is **typed** (i.e., annotated by a return type hint and/or
        one or more parameter type hints).

        Parameters
        ----------
        node : NodeCallable
            Callable node to be transformed.

        Returns
        -------
        Optional[NodeCallable]
            This same callable node.
        '''

        # If...
        if (
            # * This callable node has one or more parent nodes previously
            #   visited by this node transformer *AND* the immediate parent node
            #   of this callable node is a class node, then this callable node
            #   encapsulates a method rather than a function. In this case, the
            #   visit_ClassDef() method defined above has already explicitly
            #   decorated the class defining this method by the @beartype
            #   decorator, which then implicitly decorates both this and all
            #   other methods of that class by that decorator. For safety and
            #   efficiency, avoid needlessly re-decorating this method by the
            #   same decorator by preserving and returning this node as is.
            # * That is *NOT* the case, then this callable node is either the
            #   root node of the current AST *OR* has a parent node that is not
            #   a class node. In either case, this callable node necessarily
            #   encapsulates a function (rather than a method), which yet to be
            #   decorated. Do so now! So say we all.
            #
            # This logic corresponds to the above "That is *NOT* the case" case
            # (i.e., this callable node necessarily encapsulates a function).
            # Look. Just accept that we have a tenuous grasp on reality at best.
            not self._scopes.is_scope_class and
            # ...and the currently visited callable is annotated by one or more
            # type hints and thus *NOT* ignorable with respect to beartype
            # decoration...
            is_node_callable_typed(node)
        ):
            # print(f'Decorating function {node.name}()...')

            # Add a new child decoration node to this parent callable node
            # decorating this callable by @beartype under this configuration.
            self._decorate_node_beartype(node=node, conf=self._conf)
        # Else, that callable is ignorable. In this case, avoid needlessly
        # decorating that callable by @beartype for efficiency.

        # Recursively transform *ALL* child nodes of this parent callable node.
        return self.generic_visit(node)


    # Efficiently decorate coroutines (i.e., asynchronous callables declared via
    # the "async" keyword) by aliasing the existing visit_FunctionDef() method
    # defined above (which directly handles *ONLY* synchronous callables) to a
    # new visit_AsyncFunctionDef (which directly handles *ONLY* asynchronous
    # callables). Since the visit_FunctionDef() implementation transparently
    # handles both synchronous and asynchronous callables, this is the optimally
    # efficient approach to decorate coroutines *WITHOUT* adding any additional
    # stack frames to the call stack.
    visit_AsyncFunctionDef = visit_FunctionDef



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype decorator :pep:`525`-compliant asynchronous generator unit tests.

This submodule unit tests :pep:`525` support for asynchronous generators
implemented in the :func:`beartype.beartype` decorator.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
from beartype_test._util.mark.pytmark import ignore_warnings

# ....................{ TESTS                              }....................
# Prevent pytest from capturing and displaying all expected non-fatal
# beartype-specific warnings emitted by the @beartype decorator below. Hurk!
@ignore_warnings(BeartypeDecorHintPep585DeprecationWarning)
async def test_decor_pep525_async_generator_check() -> None:
    '''
    Test :func:`beartype.beartype`-decorated asynchronous generators with
    respect to core type-checking.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from asyncio import sleep
    from beartype import beartype
    from beartype.roar import (
        BeartypeCallHintParamViolation,
        BeartypeDecorHintPep484585Exception,
    )
    from beartype_test._util.pytroar import raises_uncached
    from collections.abc import (
        AsyncGenerator as Pep585AsyncGenerator,
        AsyncIterable as Pep585AsyncIterable,
        AsyncIterator as Pep585AsyncIterator,
    )
    from inspect import isasyncgenfunction
    from typing import (
        Union,
        AsyncGenerator as Pep484AsyncGenerator,
        AsyncIterable as Pep484AsyncIterable,
        AsyncIterator as Pep484AsyncIterator,
    )

    # ....................{ LOCALS                         }....................
    # Tuple of all PEP 484- or 585-compliant type hints annotating the returns
    # of asynchronous generators -- exercising all possible edge cases.
    HINTS_RETURN = (
        # ....................{ NON-PEP                    }....................
        # The unsubscripted "collections.abc.AsyncGenerator" superclass is a
        # PEP-noncompliant hint.
        Pep585AsyncGenerator,

        # ....................{ PEP 484                    }....................
        Pep484AsyncGenerator[Union[str, float], None],
        Pep484AsyncIterable[Union[str, float]],
        Pep484AsyncIterator[Union[str, float]],

        # ....................{ PEP 585                    }....................
        Pep585AsyncGenerator[Union[str, float], None],
        Pep585AsyncIterable[Union[str, float]],
        Pep585AsyncIterator[Union[str, float]],
    )

    # ....................{ PASS                           }....................
    # For each return type hint defined above...
    for hint_return in HINTS_RETURN:
        @beartype
        async def some_kind_of_spiritual_thing(
            said_the: Union[str, int], bigger_greener_bat: Union[str, float]
        ) -> hint_return:
            '''
            :func:`beartype.beartype`-decorated asynchronous generator whose
            return is annotated by a :pep:`484`- or :pep:`585`-compliant type
            hint deeply type-checking the value yielded by this generator.
            '''

            # Silently reduce to an asynchronous noop. See above for details.
            await sleep(0)

            # Return an arbitrary combination of the passed parameters.
            yield said_the + bigger_greener_bat

        # Assert that the high-level wrapper generated by the @beartype
        # decorator type-checking the low-level asynchronous generator defined
        # above preserves "inspect"-based metadata describing this generator.
        # Specifically, assert that this wrapper is still a generator.
        assert isasyncgenfunction(some_kind_of_spiritual_thing) is True

        # Assert that asynchronously iterating the asynchronous generator
        # created by passing this function valid parameters yields the expected
        # items. For safety, we intentionally iterate over *ALL* items of this
        # asynchronous generator -- of which there is exactly one. Although
        # asynchronous generators are technically iterators and thus support
        # manual iteration via the aiter() and anext() builtins, these builtins
        # are mostly *NOT* intended to be called manually. Whereas synchronous
        # generators *CAN* be (and often are) manually iterated by calling
        # iter() and next(), asynchronous generators technically also *CAN* be
        # (but pragmatically *NEVER* are) manually iterated by similarly calling
        # aiter() and anext(). Thus, aiter() and anext() are useless for most
        # practical intents and purposes. Whereas synchronous generators do
        # *NOT* explicitly require finalization, asynchronous generators do.
        # Why? Because asynchronous generators are registered with a parent
        # "asyncio"-based event loop and *MUST* be cleanly closed before being
        # garbage-collected. If this is *NOT* done, then Python raises obscure
        # "asyncio"-specific exceptions resembling:
        #     RuntimeWarning: coroutine method 'aclose' of
        #     'test_decor_async_generator.<locals>.some_kind_of_spiritual_thing'
        #     was never awaited
        #
        # Pytest then responds with an ever *MORE* obscure exception resembling:
        #     pytest.PytestUnraisableExceptionWarning: Exception ignored while
        #     finalizing async generator <async_generator object
        #     test_decor_async_generator.<locals>.some_kind_of_spiritual_thing
        #     at 0x7f4f959d1020>: None
        #
        # See also this relevant StackOverflow post:
        #     https://stackoverflow.com/a/42561322/2809027
        async for some_kind_of_spiritual in some_kind_of_spiritual_thing(
            'I should be trying to do some kind of spiritual thing ',
            'involving radical acceptance and enlightenment and such.',
        ):
            assert some_kind_of_spiritual == (
                'I should be trying to do some kind of spiritual thing '
                'involving radical acceptance and enlightenment and such.'
            )

        # Assert that passing this function invalid parameters raises the
        # expected violation.
        with raises_uncached(BeartypeCallHintParamViolation):
            await anext(some_kind_of_spiritual_thing(
                b'A thousand stars of sertraline ',
                b'whirled round quetiapine moons'
            ))

    # ....................{ FAIL                           }....................
    # Assert that @beartype raises the expected exception when decorating an
    # asynchronous generator annotating its return as anything *EXCEPT*
    # "AsyncGenerator[...]", "AsyncIterable[...]", or "AsyncIterator[...]".
    with raises_uncached(BeartypeDecorHintPep484585Exception):
        @beartype
        async def upside_down_trees(
            roots_in_the_breeze: str, branches_underground: str) -> str:
            # Silently reduce to an asynchronous noop. See above for details.
            await sleep(0)

            # Return an arbitrary object (sorta) satisfying this return hint.
            yield roots_in_the_breeze + branches_underground


async def test_decor_pep525_async_generator_api() -> None:
    '''
    Test :func:`beartype.beartype`-decorated asynchronous generators with
    respect to **bidirectional communication** (i.e., the pure-Python
    implementation of a hypothetical ``async yield from`` expression dynamically
    injected into the bodies of higher-level wrapper functions type-checking
    lower-level :func:`beartype.beartype`-decorated asynchronous generator
    factories).

    This unit test validates that higher-level wrapper functions type-checking
    lower-level :func:`beartype.beartype`-decorated asynchronous generator
    factories continue to comply with :pep:`525` semantics, despite the
    syntactic lack of a ``async yield from`` expression comparable to the
    existing :pep:`380`-compliant ``yield from`` expression. Specifically, this
    unit test validates comformance with:

    * :meth:`types.AsyncGeneratorType.aclose` semantics.
    * :meth:`types.AsyncGeneratorType.asend` semantics.
    * :meth:`types.AsyncGeneratorType.athrow` semantics.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from asyncio import sleep
    from beartype import beartype
    from collections.abc import AsyncGenerator
    from pytest import raises

    # ....................{ EXCEPTIONS                     }....................
    # Arbitrary list to be appended to by the "finally:" block of the non-empty
    # asynchronous generator why_do_i_know_ye() defined below, enabling logic
    # below to validate that that block was run as expected *AFTER* the caller
    # prematurely closes that generator by calling its aclose() method.
    to_see_and_to_behold = []

    # Arbitrary string to be inserted into this list below.
    HORRORS_NEW = 'To see and to behold these horrors new?'

    # ....................{ EXCEPTIONS                     }....................
    class _BeartypeAsyncGeneratorThrowException(Exception):
        '''
        Arbitrary exception asynchronously thrown below into an asynchronous
        generator during iteration.
        '''

        pass

    # ....................{ CALLABLES                      }....................
    @beartype
    async def is_my_eternal_essence() -> AsyncGenerator:
        '''
        :func:`beartype.beartype`-decorated asynchronous generator guaranteed to
        finalize *before* yielding anything.

        See Also
        --------
        https://stackoverflow.com/a/13243870/2809027
            StackOverflow answer inspiring this implementation.
        '''

        # Silently reduce to an asynchronous noop. See above for details.
        await sleep(0)

        # Immediately return *BEFORE* yielding. (Look. Don't ask.)
        return

        # Declare this function to be an asynchronous generator factory rather
        # than an asynchronous coroutine. Without this "yield" statement, this
        # function would instead be interpreted as the latter.
        #
        # This asynchronous generator is guaranteed to return before yielding.
        # This asynchronous generator is guaranteed to thus be empty.
        yield


    @beartype
    async def why_do_i_know_ye(yield_int_max: int) -> AsyncGenerator[int, str]:
        '''
        :func:`beartype.beartype`-decorated non-empty asynchronous generator
        yielding:

        * When sent an arbitrary string, the length of that string.
        * Else, non-negative integers in the inclusive range
          ``[0, yield_int_max]``.

        Parameters
        ----------
        yield_int_max : int
            Maximum non-negative integer to be yielded.
        '''
        assert yield_int_max >= 0

        # Silently reduce to an asynchronous noop. See above for details.
        await sleep(0)

        # Attempt to...
        try:
            # For each integer in this range to be yielded...
            for yield_int in range(yield_int_max + 1):
                # Attempt to...
                try:
                    # Yield this integer to the caller *AND* possibly capture a
                    # string sent by the caller into this generator.
                    sent_str = yield yield_int

                    # While the caller continues sending strings into this
                    # generator, yield the length of each such string as
                    # confirmation of receipt to the caller.
                    while sent_str is not None:
                        sent_str = yield len(sent_str)
                    # Else, the caller sent *NO* strings into this generator.
                # If the caller explicitly threw an exception into this
                # generator by calling this generator's athrow() method, yield
                # the negation of the length of this exception's message as
                # confirmation of receipt.
                except _BeartypeAsyncGeneratorThrowException as exception:
                    yield -len(str(exception))  # <-- wat lol
        # Append an arbitrary object to the list defined above, enabling
        # logic below to validate this finalizer was run as expected after
        # the caller prematurely closes this generator with aclose().
        finally:
            to_see_and_to_behold.append(HORRORS_NEW)

    # ....................{ ASSERTS ~ empty : yield        }....................
    # Validate that a @beartype-decorated empty asynchronous generator factory
    # preserves PEP 580-compliant "yield" semantics.

    # True only if this generator truly is empty.
    is_my_eternal_essence_empty = True

    # If this generator yields *ANY* value, note this generator to be non-empty.
    async for yielded_value in is_my_eternal_essence():
        is_my_eternal_essence_empty = False

    # Assert that this generator truly is empty.
    assert is_my_eternal_essence_empty is True

    # ....................{ ASSERTS ~ non-empty : yield    }....................
    # Validate that a @beartype-decorated non-empty asynchronous generator
    # factory preserves PEP 580-compliant "yield" semantics.

    # List of all values subsequently yielded by iterating over this generator.
    yielded_values = []

    # Iteratively append each value yielded by this generator when passed a
    # small (but still non-zero) maximum value to be yielded.
    async for yielded_value in why_do_i_know_ye(3):
        yielded_values.append(yielded_value)

    # Assert that this generator yielded the expected values.
    assert yielded_values == [0, 1, 2, 3]

    # Assert that this generator's "finally:" block finalized this generator by
    # appending the expected string to this closure list.
    assert to_see_and_to_behold == [HORRORS_NEW]

    # Clear this list for sanity.
    to_see_and_to_behold.clear()

    # ....................{ ASSERTS ~ non-empty : send     }....................
    # Validate that a @beartype-decorated non-empty asynchronous generator
    # factory preserves PEP 580-compliant asend() semantics.

    # Asynchronous generator produced by priming this factory as above.
    why_have_i_seen_ye = why_do_i_know_ye(3)
    # print(f'why_have_i_seen_ye: {dir(why_have_i_seen_ye)}')

    # Assert that the first value yielded by this generator is the expected.
    # Rather than iterate this generator manually via the standard idiom,
    # validate that sending the "None" singleton into this generator is
    # semantically equivalent to manually iterating this generator.
    #
    # Note that:
    # * This obtuse alternative to explicit iteration is equivalent to this
    #   standard idiom for manually iterating an asynchronous generator:
    #     assert await anext(why_have_i_seen_ye) == 0
    # * One *CANNOT* send a non-"None" value into an asynchronous generator
    #   before that generator yields its first value after being iterated at
    #   least once. Violating that maxim raises this exception:
    #     TypeError: can't send non-None value to a just-started async generator
    assert await why_have_i_seen_ye.asend(None) == 0

    # Assert that the next value yielded by this generator is the expected.
    assert await anext(why_have_i_seen_ye) == 1

    # Assert that the value yielded by this generator after sending a string
    # into this generator is the length of this string.
    assert await why_have_i_seen_ye.asend(
        'Why do I know ye? why have I seen ye? why') == 41

    # Assert that the value yielded by this generator after sending another
    # string into this generator is the length of this string.
    assert await why_have_i_seen_ye.asend(
        'Is my eternal essence thus distraught') == 37

    # Assert that the next value yielded by this generator is the expected.
    assert await anext(why_have_i_seen_ye) == 2

    # ....................{ ASSERTS ~ non-empty : throw    }....................
    # Assert that the value yielded by this generator after throwing an
    # exception into this generator is the negation of the length of this
    # exception's message. Just accept it.
    assert await why_have_i_seen_ye.athrow(
        _BeartypeAsyncGeneratorThrowException(
            'To see and to behold these horrors new?')) == -39

    # ....................{ ASSERTS ~ non-empty : send END }....................
    # Assert that the final value yielded by this generator is the expected.
    assert await anext(why_have_i_seen_ye) == 3

    # Assert that iterating this generator one final time raises the expected
    # PEP 525-compliant exception -- implying this generator has nothing further
    # to yield and has thus been exhausted (i.e., finalized, closed).
    with raises(StopAsyncIteration):
        await anext(why_have_i_seen_ye)

    # Assert that this generator's "finally:" block finalized this generator by
    # appending the expected string to this closure list.
    assert to_see_and_to_behold == [HORRORS_NEW]

    # Clear this list for sanity.
    to_see_and_to_behold.clear()

    # ....................{ ASSERTS ~ non-empty : close    }....................
    # Asynchronous generator produced by priming this factory as above.
    thus_distraught = why_do_i_know_ye(3)

    # Assert that the first value yielded by this generator is the expected.
    assert await anext(thus_distraught) == 0

    # Prematurely close this generator.
    await thus_distraught.aclose()

    # Assert that this generator's "finally:" block finalized this generator by
    # appending the expected string to this closure list.
    assert to_see_and_to_behold == [HORRORS_NEW]

    # Clear this list for sanity.
    to_see_and_to_behold.clear()

    # Assert that iterating this generator one final time raises the expected
    # PEP 525-compliant exception -- implying this generator has nothing further
    # to yield and has thus been exhausted (i.e., finalized, closed).
    with raises(StopAsyncIteration):
        await anext(thus_distraught)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide SQLAlchemy integration tests.

This submodule functionally tests the :mod:`beartype` package against the
third-party SQLAlchemy package.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytskip import skip_unless_package

# ....................{ TESTS                              }....................
@skip_unless_package('sqlalchemy')
def test_sqlalchemy_asyncsession() -> None:
    '''
    Integration test validating that the :mod:`beartype` package raises *no*
    unexpected exceptions when type-checking instances of the third-party
    :class:`sqlalchemy.ext.asyncio.AsyncSession` generic, whose non-triviality
    once caused :mod:`beartype` to raise spurious type-checking violations.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.door import die_if_unbearable
    from beartype.typing import Optional
    from sqlalchemy.ext.asyncio import AsyncSession

    # ....................{ ASSERTS                        }....................
    # Assert that @beartype at least superficially supports this non-trivial
    # generic by generating type-checking code *WITHOUT* raising spurious
    # exceptions. Ideally, this test would fully instantiate an asynchronous
    # SQLALchemy session and pass that as well. Pragmatically, doing so is
    # predictably non-trivial and altogether not worth the hassle. KISS it, yo!
    die_if_unbearable(None, Optional[AsyncSession])



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **type hint factories** (i.e., PEP-compliant type hint factories
supplementing standard type hint factories published by the :mod:`typing` module
with custom behaviours, including backward compatibility with older Python
versions that would otherwise *not* support those factories).

This private submodule is intentionally distinct from the lower-level private
:data:`beartype._data.typing.datatyping` submodule.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# CAUTION: This submodule *CANNOT* import from the companion "datatyping"
# submodule due to circular import dependencies between these two submodules.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

from beartype.typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    FrozenSet,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
)
from beartype._data.kind.datakindiota import Iota
from beartype._util.api.standard.utiltyping import (
    import_typing_attr_or_fallback)
from beartype._util.hint.utilhintfactory import TypeHintTypeFactory
# from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_11

# Note that, although this higher-level submodule can safely import from the
# lower-level "datatyping" submodule, the reverse is *NOT* the case.
from beartype._data.typing.datatyping import (
    Pep484612646TypeArgUnpacked)

# ....................{ FACTORIES                          }....................
#FIXME: This approach is *PHENOMENAL.* No. Seriously, We could implement a
#full-blown "beartype.typing" subpackage (or perhaps even separate "beartyping"
#package) extending this core concept to *ALL* type hint factories, enabling
#users to trivially annotate with any type hint factory regardless of the
#current version of Python or whether "typing_extensions" is installed or not.

# Portably backport *ALL* type hint factories introduced by subsequent Python
# interpreters more recent than that of the oldest actively maintained Python
# interpreter still supported by @beartype, regardless of the current version of
# Python and regardless of whether this submodule is currently being subject to
# static type-checking or not. Praise be to MIT ML guru and stunning Hypothesis
# maintainer @rsokl (Ryan Soklaski) for this brilliant circumvention. \o/
#
# If this submodule is currently being statically type-checked (e.g., mypy),
# intentionally import from the third-party "typing_extensions" module rather
# than the standard "typing" module. Why? Because doing so eliminates Python
# version complaints from static type-checkers (e.g., mypy, pyright). Static
# type-checkers could care less whether "typing_extensions" is actually
# installed or not; they only care that "typing_extensions" unconditionally
# defines this type factory across all Python versions, whereas "typing" only
# conditionally defines these type factories under particular Python versions.
if TYPE_CHECKING:
    #FIXME: Actually, this now seems to reduce to a noop. Excise us up, please.
    # # See discussion below, please. *sigh*
    # from typing_extensions import TypeAlias

    # ....................{ PEP 742                        }....................
    # PEP 742-compliant "TypeIs[...]" type hints first introduced by Python
    # >= 3.13 are a high internal priority for @beartype. Hinting the return of
    # the is_bearable() tester with a type guard created by this factory
    # effectively coerces that tester into an arbitrarily "smart" type narrower
    # and thus type parser at static analysis time, reducing complaints from
    # static type-checkers in downstream code deferring to that tester.

    #FIXME: Unconditionally globalize this *AFTER* dropping Python 3.12: e.g.,
    #   from typing import TypeIs as TypeIs
    from typing_extensions import TypeIs as TypeIs

    # ....................{ PEP 747                        }....................
    # PEP 747-compliant "TypeForm[...]" type hints first introduced by Python
    # >= 3.15 are a high internal priority for @beartype. Hinting parameters,
    # returns, and local variables of both private and public @beartype
    # callables (including the common first parameter "hint" accepted by most
    # callables) with type forms created by this factory enables static
    # type-checkers to ensure that these objects are actually type hints.
    #
    # Note that we intentionally alias "TypeForm" to a more concise and readable
    # name. The term "type form" does *NOT* especially mean much within the
    # context of Python type hints. The term "hint", on the other hand, does.

    # If this static type-checker is mypy, avoid importing PEP 747-compliant
    # type hint factories. Mypy currently lacks official support for PEP 747.
    # Instead, import standard PEP 484-compliant type hint factories guaranteed
    # to be supported by mypy that trivially reduce to the "Any" noop.
    #
    # See also these relevant mypy threads:
    #     https://github.com/python/mypy/pull/18690
    #     https://github.com/python/mypy/issues/19227
    MYPY = False  # <-- don't ask, don't tell
    if MYPY:  # <------ don't see what you don't want to see
        from beartype.typing import (
            Any as Hint,
            Generic,
        )

        _T = TypeVar('_T')
        '''
        Arbitrary type variable.
        '''

        class HintBare(Generic[_T]):
            '''
            Arbitrary generic type hint factory returning the :obj:`typing.Any`
            singleton when subscripted by *any* type hint.

            This factory is intentionally defined as a generic to prevent mypy
            from emitting false positives resembling:

                beartype/door/_func/doorcheck.py:131: error: "HintBare" expects
                no type arguments, but 1 given  [type-arg]
            '''

            @classmethod
            def __class_getitem__(cls, item: Any) -> Any:
                return Any
    # Else, this static type-checker is *NOT* mypy. Since the only other static
    # type-checker officially supported by beartype is pyright, this static
    # type-checker *MUST* by definition be pyright. Since pyright officially
    # supports PEP 747, import PEP 747-compliant type hint factories.
    else:
        #FIXME: Replace "typing_extensions" with simply "typing" *AFTER*
        #dropping Python 3.14.
        from typing_extensions import (
            TypeForm as Hint,
            TypeForm as HintBare,  # pyright: ignore
        )
# Else, this submodule is currently being imported at runtime by Python. In this
# case, dynamically import these type factories from whichever of the standard
# "typing" module *OR* the third-party "typing_extensions" module declares these
# factories, falling back to various builtin types if none do.
else:
    # ....................{ PEP 742                        }....................
    TypeIs = import_typing_attr_or_fallback(
        'TypeIs', TypeHintTypeFactory(bool))

    # ....................{ PEP 747                        }....................
    TypeForm = import_typing_attr_or_fallback(
        'TypeForm', TypeHintTypeFactory(object))


    HintBare = TypeForm
    '''
    PEP-compliant type hint matching *any* PEP-compliant type hint.

    This hint should *only* be used to explicitly subscript the
    :obj:`typing.TypeForm` type hint factory. For all other purposes, the
    subscripted :data:`.Hint` type hint should be strongly preferred.
    '''


    Hint = TypeForm[Any]
    '''
    PEP-compliant type hint matching *any* PEP-compliant type hint.

    Although semantically equivalent to ``typing.TypeForm[object]``, the
    unsubscripted :obj:`typing.TypeForm` type hint factory is currently unusable
    as such under Python <= 3.14. This insane hack trivially circumvents that.
    '''

# ....................{ EXPORT                             }....................
# Explicitly export the "Hint" and "HintBare" aliases of the "typing.TypeForm"
# hint factory imported above. Blame mypy.
__all__ = [
    'Hint',
    'HintBare',
    'TypeIs',
]

# ....................{ HINTS                              }....................
HintOrNone = Optional[Hint]
'''
PEP-compliant type hint matching either any PEP-compliant type hint *or*
:data:`None`.
'''


# ....................{ HINTS ~ container                  }....................
#FIXME: Ideally, all of the below should themselves be annotated as ": Hint".
#Mypy likes that but pyright hates that. This is why we can't have good things.

FrozenSetHints = FrozenSet[Hint]
'''
:pep:`585`-compliant type hint matching *any* **type hint frozen set** (i.e.,
frozen set of zero or more type hints).
'''


IterableHints = Iterable[Hint]
'''
:pep:`585`-compliant type hint matching *any* **type hint iterable** (i.e.,
iterable iteratively yielding zero or more type hints).
'''


ListHints = List[Hint]
'''
:pep:`585`-compliant type hint matching *any* **type hint list** (i.e., list of
zero or more type hints).
'''


SequenceHints = Sequence[Hint]
'''
:pep:`585`-compliant type hint matching *any* **type hint sequence** (i.e.,
sequence of zero or more type hints).
'''


SetHints = Set[Hint]
'''
:pep:`585`-compliant type hint matching *any* **type hint set** (i.e., set of
zero or more type hints).
'''

# ....................{ HINTS ~ container : dict           }....................
DictStrToHint = Dict[str, Hint]
'''
:pep:`585`-compliant type hint matching a dictionary mapping from strings to
PEP-compliant type hints.
'''

# ....................{ HINTS ~ container : tuple          }....................
TupleHints = Tuple[Hint, ...]
'''
:pep:`585`-compliant type hint matching *any* **child type hints** (i.e., tuple
of zero or more child type hints subscripting a parent type hint).
'''


HintOrTupleHints = Union[Hint, TupleHints]
'''
:pep:`585`-compliant type hint matching either a single type hint *or* a tuple
of zero or more type hints.
'''

# ....................{ PEP ~ 484                          }....................
Pep484TypeVarToHint = Dict[TypeVar, Hint]
'''
:pep:`585`-compliant type hint matching a **type variable lookup table** (i.e.,
dictionary mapping from :pep:`484`-compliant type variables to the arbitrary
type hints those type variables map to).

Type variable lookup tables are commonly employed throughout the :mod:`beartype`
codebase to record **type variable substitutions** (i.e., the dynamic
replacement of type variables by non-type variables in larger type hints).
'''


T_Hint = TypeVar('T_Hint', bound=Hint)
'''
:pep:`484`-compliant **type hint type variable** (i.e.,
:class:`typing.TypeVar` object bound to match *only* PEP-compliant type
hints).
'''

# ....................{ PEP ~ (484|604)                    }....................
HintOrSentinel = Union[Hint, Iota]
'''
:pep:`484` and :pep:`604`-compliant union matching both PEP-compliant type hints
*and* the sentinel placeholder (i.e.,
:obj:`beartype._data.kind.datakindiota.SENTINEL` singleton).
'''

# ....................{ PEP ~ (484|646)                    }....................
Pep484612646TypeArgUnpackedToHint = Dict[Pep484612646TypeArgUnpacked, Hint]
'''
:pep:`585`-compliant type hint matching a :pep:`484`-, :pep:`612`-, and
:pep:`646`-compliant **unpacked type parameter lookup table** (i.e., dictionary
mapping from **unpacked type parameters** (i.e., :pep:`484`-compliant type
variables, :pep:`612`-compliant unpacked parameter specifications, and
:pep:`646`-compliant unpacked type variable tuples) to the arbitrary type hints
those type parameters map to).

Type parameter lookup tables are commonly employed throughout the
:mod:`beartype` codebase to record **type parameter substitutions** (i.e., the
dynamic replacement of type parameter by non-type parameter in larger type
hints). For this reason, this hint intentionally matches dictionaries whose
keys are unpacked rather than packed type parameters. Why? Because type
parameters are *always* specified in unpacked rather than packed form. Packed
type parameters are thus useless for most intents and purposes.

For example, in the generic class declaration:

* ``class DisIsATensorYo[*Ts]()``, the ``*Ts`` denotes a :pep:`646`-compliant
  unpacked type variable tuple.
* ``class DisIsNotATensorYo[T]()``, the ``T`` denotes a :pep:`484`-compliant
  type variable.

It's literally infeasible to syntactically subscript or parametrize a type hint
by a :pep:`646`-compliant packed type variable tuple.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **Click utilities** (i.e., low-level callables handling the
third-party :mod:`click` package).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# CAUTION: The top-level of this module should avoid importing from third-party
# optional libraries, both because those libraries cannot be guaranteed to be
# either installed or importable here *AND* because those imports are likely to
# be computationally expensive, particularly for imports transitively importing
# C extensions (e.g., anything from NumPy or SciPy).
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype.roar import BeartypeDecorWrappeeException
from beartype._data.typing.datatyping import BeartypeableT

# ....................{ DECORATORS                         }....................
def beartype_click_command(
    click_command: BeartypeableT, **kwargs) -> BeartypeableT:
    '''
    Decorate the passed **Click command** (i.e., object produced by the
    :func:`click.command` decorator decorating an arbitrary callable) with
    dynamically generated type-checking.

    Design
    ------
    Note that the :func:`click.command` decorator does *not* support standard
    decorator chaining "out of the box." Click commands thus require manual
    detection and intervention from the :func:`beartype.beartype` decorator.
    Previously, we attempted to generically decorate Click commands using
    standard decorator chaining; that is to say, we preferred to ignore the
    existence of Click commands. Sadly, doing so caused Click to raise
    unreadable exceptions resembling the following -- indicative of low-level
    unresolved issues within Click itself:

    .. code-block:: python

       >>> from beartype import beartype
       >>> from click import command
       >>> from click.testing import CliRunner

       >>> @beartype
       ... @command()
       ... def main() -> None: pass

       >>> runner = CliRunner()
       >>> result = runner.invoke(cli=main)
       Traceback (most recent call last):
         File "/usr/lib/python3.13/site-packages/click/testing.py", line 403, in invoke
           prog_name = extra.pop("prog_name")
       KeyError: 'prog_name'

       During handling of the above exception, another exception occurred:

       Traceback (most recent call last):
         File "/home/leycec/tmp/mopy.py", line 19, in <module>
           result = runner.invoke(cli=main)
         File "/usr/lib/python3.13/site-packages/click/testing.py", line 405, in invoke
           prog_name = self.get_default_prog_name(cli)
         File "/usr/lib/python3.13/site-packages/click/testing.py", line 195, in get_default_prog_name
           return cli.name or "root"
                  ^^^^^^^^
       AttributeError: 'NoneType' object has no attribute 'name'

    Parameters
    ----------
    func : BeartypeableT
        Click command to be decorated by :func:`beartype.beartype`.

    All remaining keyword parameters are passed as is to the lower-level
    :func:`._beartype_func` decorator internally called by this higher-level
    decorator on the pure-Python function encapsulated in this Click command.

    Returns
    -------
    BeartypeableT
        New pure-Python callable wrapping this Click command with type-checking.
    '''

    # Avoid circular and third-party import dependencies.
    from beartype._decor._nontype.decornontype import beartype_func
    from click.core import Command  # pyright: ignore

    # If this Click command is *NOT* actually a @click.command()-decorated
    # callable, raise an exception.
    if not isinstance(click_command, Command):
        raise BeartypeDecorWrappeeException(  # pragma: no cover
            f'Click command {repr(click_command)} not  '
            f'decorated by @click.command().'
        )
    # Else, this Click command is a @click.command()-decorated callable.

    # Old pure-Python callable decorated by the @click.command() decorator.
    func = click_command.callback  # pyright: ignore

    # New pure-Python callable decorating that callable with type-checking.
    func_checked = beartype_func(func=func, **kwargs)  # type: ignore[type-var]

    # Replace the old with new pure-Python callable in the Click command created
    # and returned by the @click.command() decorator.
    click_command.callback = func_checked  # pyright: ignore

    # Return the same Click command.
    return click_command  # type: ignore[return-value]



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **abstract syntax tree (AST) mungers** (i.e., low-level callables
modifying various properties of various nodes in the currently visited AST).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from ast import (
    AST,
    Call,
    ClassDef,
    Name,
    Subscript,
    expr,
    keyword,
)
from beartype._data.api.standard.dataast import NODE_CONTEXT_LOAD
from beartype._data.claw.dataclawmagic import BEARTYPE_CLAW_STATE_OBJ_NAME
from beartype._util.ast.utilastmake import (
    make_node_kwarg,
    make_node_object_attr_load,
    make_node_str,
)
from beartype._util.ast.utilastmunge import copy_node_metadata

# ....................{ SUBCLASSES                         }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# CAUTION: To improve forward compatibility with the superclass API over which
# we have *NO* control, avoid accidental conflicts by suffixing *ALL* private
# and public attributes of this subclass by "_beartype".
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

class BeartypeNodeTransformerUtilityMixin(object):
    '''
    **Beartype abstract syntax tree (AST) node utility transformer** (i.e.,
    low-level mixin of the high-level
    :class:`beartype.claw._ast.clawastmain.BeartypeNodeTransformer` subclass
    supplementing that subclass with various low-level methods creating,
    modifying, and introspecting common node types and subclass properties).
    '''

    # ....................{ PRIVATE ~ factories            }....................
    #FIXME: Unit test us up, please.
    def _make_node_keyword_conf(self, node_sibling: AST) -> keyword:
        '''
        Create and return a new **beartype configuration keyword argument node**
        (i.e., abstract syntax tree (AST) node passing the beartype
        configuration associated with the currently visited module as a ``conf``
        keyword to a :func:`beartype.beartype` decorator orchestrated by the
        caller).

        Parameters
        ----------
        node_sibling : AST
            Sibling node to copy source code metadata from.

        Returns
        -------
        keyword
            Keyword node passing this configuration to an arbitrary function.
        '''

        # Node encapsulating the fully-qualified name of the current module.
        node_module_name = make_node_str(
            text=self._module_name, node_sibling=node_sibling)  # type: ignore[attr-defined]

        # Node encapsulating a reference to the beartype configuration object
        # cache (i.e., dictionary mapping from fully-qualified module names to
        # the beartype configurations associated with those modules).
        node_module_name_to_conf = make_node_object_attr_load(
            obj_name=BEARTYPE_CLAW_STATE_OBJ_NAME,
            attr_name='module_name_to_beartype_conf',
            node_sibling=node_sibling,
        )

        # Expression node encapsulating the indexation of a dictionary by the
        # fully-qualified name of the current module. For simplicity, simply
        # reuse this node.
        node_module_name_index = node_module_name

        # Node encapsulating a reference to this beartype configuration,
        # indirectly (and efficiently) accessed via a dictionary lookup into
        # this object cache. While cumbersome, this indirection is effectively
        # "glue" integrating this AST node generation algorithm with the
        # corresponding Python code subsequently interpreted by Python at
        # runtime during module importation.
        node_conf = Subscript(
            value=node_module_name_to_conf,
            slice=node_module_name_index,  # type: ignore[arg-type]
            ctx=NODE_CONTEXT_LOAD,
        )

        # Node encapsulating the passing of this beartype configuration as the
        # "conf" keyword argument to an arbitrary function call of some suitable
        # "beartype" function orchestrated by the caller.
        node_keyword_conf = make_node_kwarg(
            kwarg_name='conf', kwarg_value=node_conf, node_sibling=node_sibling)

        # Copy all source code metadata (e.g., line numbers) from this sibling
        # node onto these new nodes.
        copy_node_metadata(node_src=node_sibling, node_trg=node_conf)

        # Return this "conf" keyword node.
        return node_keyword_conf



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Callable argument tester utility unit tests.**

This submodule unit tests the public API of the private
:mod:`beartype._util.func.arg.utilfuncargtest` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ kind                       }....................
def test_is_func_argless() -> None:
    '''
    Test the
    :func:`beartype._util.func.arg.utilfuncargtest.is_func_argless` tester.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.func.arg.utilfuncargtest import is_func_argless
    from beartype_test.a00_unit.data.func.data_func import (
        func_args_0,
        func_args_1_flex_mandatory,
        func_args_1_varpos,
        func_args_1_kwonly_mandatory,
    )

    # ....................{ ASSERTS                        }....................
    # Assert this tester accepts an argumentless callable.
    assert is_func_argless(func_args_0) is True

    # Assert this tester rejects all other callables.
    assert is_func_argless(func_args_1_flex_mandatory) is False
    assert is_func_argless(func_args_1_varpos) is False
    assert is_func_argless(func_args_1_kwonly_mandatory) is False


def test_is_func_arg_variadic() -> None:
    '''
    Test the
    :func:`beartype._util.func.arg.utilfuncargtest.is_func_arg_variadic`
    tester.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.func.arg.utilfuncargtest import is_func_arg_variadic

    # ....................{ CALLABLES                      }....................
    #FIXME: Generalize these callables into the
    #"beartype_test.a00_unit.data.func.data_func" submodule, please.

    # Arbitrary callable accepting *NO* variadic arguments.
    def by_day_or_starlight(thus_from: int, my_first_dawn: int) -> int:
        return thus_from + my_first_dawn

    # Arbitrary callable accepting a variadic positional argument.
    def of_childhood(
        didst_thou: str, *args: str, intertwine_for_me: str) -> str:
        return didst_thou + ''.join(args) + intertwine_for_me

    # Arbitrary callable accepting a variadic keyword argument.
    def the_passions(
        that_build_up: tuple, our_human_soul: tuple, **kwargs: tuple) -> tuple:
        return that_build_up + our_human_soul + (
            kwarg_tuple_item
            for kwarg_tuple in kwargs
            for kwarg_tuple_item in kwarg_tuple
        )

    # ....................{ ASSERTS                        }....................
    # Assert this tester accepts callables accepting variadic arguments.
    assert is_func_arg_variadic(of_childhood) is True
    assert is_func_arg_variadic(the_passions) is True

    # Assert this tester rejects callables accepting *NO* variadic arguments.
    assert is_func_arg_variadic(by_day_or_starlight) is False

# ....................{ TESTS ~ name                       }....................
def test_is_func_arg_name() -> None:
    '''
    Test the
    :func:`beartype._util.func.arg.utilfuncargtest.is_func_arg_name` tester.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.func.arg.utilfuncargtest import is_func_arg_name
    from beartype_test.a00_unit.data.func.data_func import (
        func_args_5_flex_mandatory_varpos_kwonly_varkw)

    # ....................{ ASSERTS                        }....................
    # Assert this tester accepts all argument names accepted by the passed
    # callable.
    assert is_func_arg_name(
        func=func_args_5_flex_mandatory_varpos_kwonly_varkw,
        arg_name='we_are_selfish_men',
    ) is True
    assert is_func_arg_name(
        func=func_args_5_flex_mandatory_varpos_kwonly_varkw,
        arg_name='oh_raise_us_up',
    ) is True
    assert is_func_arg_name(
        func=func_args_5_flex_mandatory_varpos_kwonly_varkw,
        arg_name='return_to_us_again',
    ) is True
    assert is_func_arg_name(
        func=func_args_5_flex_mandatory_varpos_kwonly_varkw,
        arg_name='and_give_us',
    ) is True
    assert is_func_arg_name(
        func=func_args_5_flex_mandatory_varpos_kwonly_varkw,
        arg_name='manners_virtue_freedom_power',
    ) is True

    # Assert this tester rejects all local variable names declared in the body
    # of the passed callable.
    assert is_func_arg_name(
        func=func_args_5_flex_mandatory_varpos_kwonly_varkw,
        arg_name='thy_soul_was_like_a_star',
    ) is False


def test_is_func_arg_name_is_func_arg_name_variadic_positional() -> None:
    '''
    Test the
    :func:`beartype._util.func.arg.utilfuncargtest.is_func_arg_name_variadic_positional`
    tester.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.func.arg.utilfuncargtest import (
        is_func_arg_name_variadic_positional)
    from beartype_test.a00_unit.data.func.data_func import (
        func_args_3_flex_mandatory_optional_varkw_wrapped)
    from beartype_test.a00_unit.data.func.data_pep570 import (
        func_args_10_all_except_flex_mandatory)

    # ....................{ ASSERTS                        }....................
    # Assert that this tester rejects a callable accepting a parameter with the
    # passed name that is *NOT* a variadic positional parameter.
    assert is_func_arg_name_variadic_positional(
        func=func_args_3_flex_mandatory_optional_varkw_wrapped,
        arg_name='rude_bare_and_high',
    ) is False

    # Assert that this tester rejects a callable accepting a variadic positional
    # parameter with a different name than that of the passed name.
    assert is_func_arg_name_variadic_positional(
        func=func_args_10_all_except_flex_mandatory,
        arg_name='rain_is_pouring_down',
    ) is False

    # Assert that this tester rejects a callable accepting a variadic positional
    # parameter but accepting *NO* parameter with the passed name.
    assert is_func_arg_name_variadic_positional(
        func=func_args_10_all_except_flex_mandatory,
        arg_name='before_spreading_his_black_wings',
    ) is False

    # Assert that this tester accepts a callable accepting a variadic positional
    # parameter with the passed name.
    assert is_func_arg_name_variadic_positional(
        func=func_args_10_all_except_flex_mandatory,
        arg_name='in_the_rain_my_tears_are_forever_lost',
    ) is True


def test_is_func_arg_name_is_func_arg_name_variadic_keyword() -> None:
    '''
    Test the
    :func:`beartype._util.func.arg.utilfuncargtest.is_func_arg_name_variadic_keyword`
    tester.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.func.arg.utilfuncargtest import (
        is_func_arg_name_variadic_keyword)
    from beartype_test.a00_unit.data.func.data_func import (
        func_args_1_varpos_wrapped)
    from beartype_test.a00_unit.data.func.data_pep570 import (
        func_args_10_all_except_flex_mandatory)

    # ....................{ ASSERTS                        }....................
    # Assert that this tester rejects a callable accepting a parameter with the
    # passed name that is *NOT* a variadic keyword parameter.
    assert is_func_arg_name_variadic_keyword(
        func=func_args_1_varpos_wrapped,
        arg_name='and_in_her_his_one_delight',
    ) is False

    # Assert that this tester rejects a callable accepting a variadic keyword
    # parameter with a different name than that of the passed name.
    assert is_func_arg_name_variadic_keyword(
        func=func_args_10_all_except_flex_mandatory,
        arg_name='rain_is_pouring_down',
    ) is False

    # Assert that this tester rejects a callable accepting a variadic keyword
    # parameter but accepting *NO* parameter with the passed name.
    assert is_func_arg_name_variadic_keyword(
        func=func_args_10_all_except_flex_mandatory,
        arg_name='before_spreading_his_black_wings',
    ) is False

    # Assert that this tester accepts a callable accepting a variadic keyword
    # parameter with the passed name.
    assert is_func_arg_name_variadic_keyword(
        func=func_args_10_all_except_flex_mandatory,
        arg_name='sitting_in_calmness',
    ) is True



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **type-checking function code snippets** (i.e., triple-quoted
pure-Python string constants formatted and concatenated together to dynamically
generate the implementations of functions type-checking arbitrary objects
against arbitrary PEP-compliant type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.code.datacodename import (
    ARG_NAME_CHECK_META,
    ARG_NAME_CONF,
    ARG_NAME_EXCEPTION_PREFIX,
    ARG_NAME_GET_VIOLATION,
    ARG_NAME_HINT,
    ARG_NAME_WARN,
    VAR_NAME_PITH_ROOT,
    VAR_NAME_RANDOM_INT,
    VAR_NAME_VIOLATION,
)

# ....................{ CODE ~ signature                   }....................
CODE_CHECKER_SIGNATURE = f'''{{code_signature_prefix}}def {{func_name}}(
    {VAR_NAME_PITH_ROOT},
{{code_signature_scope_args}}
):'''
'''
Code snippet declaring the signature of all type-checking tester functions
created by the :func:`beartype._data.code.datacodename.make_func_tester` factory.

Note that:

* This signature intentionally:

  * Avoids annotating its parameters or return by type hints. Doing so would be:

    * Pointless, as the type-checking functions dynamically created and returned
      by factory functions defined by the "beartype._check.checkmake" submodule
      are only privately called by the public beartype.door.is_bearable() and
      beartype.door.die_if_unbearable() runtime type-checkers.
    * Harmful, as doing so would prevent this common signature from being
      generically reused as the signature for both raisers and testers.

  * Names the single public parameter accepted by this tester function
    ``{VAR_NAME_PITH_ROOT}``. Doing so trivially ensures that the memoized
    type-checking boolean expression generated by the
    :func:`beartype._check.code.codemain.make_check_expr` code factory
    implicitly type-checks the passed object *without* further modification
    (e.g., global search-and-replacement), ensuring that memoized expression may
    be efficiently reused as is *without* subsequent unmemoization. Clever, huh?

* ``code_signature_prefix`` is usually either:

  * For synchronous callables, the empty string.
  * For asynchronous callables (e.g., asynchronous generators, coroutines), the
    space-suffixed keyword ``"async "``.
'''

# ....................{ CODE ~ check                       }....................
CODE_TESTER_CHECK_PREFIX = '''
    # Return true only if the passed object satisfies this type hint.
    return '''
'''
Code snippet prefixing the type-check of an arbitrary object passed to a
type-checking tester function against an arbitrary type hint passed to the same
function.
'''

# ....................{ CODE ~ check                       }....................
CODE_RAISER_HINT_OBJECT_CHECK_PREFIX = '''

    # Type-check this object against this type hint.
    if not '''
'''
Code snippet prefixing the type-check of an arbitrary object passed to a
type-checking raiser function against an arbitrary type hint passed to the same
function.
'''


CODE_RAISER_FUNC_PITH_CHECK_PREFIX = '''
        # Type-check this parameter or return against this type hint.
        if not '''
'''
Code snippet prefixing the type-check of a parameter or return of a decorated
callable against the type hint annotating that parameter or return.
'''

# ....................{ CODE ~ violation : get             }....................
CODE_GET_HINT_OBJECT_VIOLATION = f''':
            {VAR_NAME_VIOLATION} = {ARG_NAME_GET_VIOLATION}(
                obj={VAR_NAME_PITH_ROOT},
                hint={ARG_NAME_HINT},
                conf={ARG_NAME_CONF},
                exception_prefix={ARG_NAME_EXCEPTION_PREFIX},{{arg_random_int}}
            )
'''
'''
Code snippet suffixing all code type-checking the **root pith** (i.e., arbitrary
object) against the root type hint annotating that pith by either raising a
fatal exception or emitting a non-fatal warning.

This snippet expects to be formatted with these named interpolations:

* ``{arg_random_int}``, whose value is either:

  * If type-checking for the current type hint requires a pseudo-random integer,
    :data:`.CODE_HINT_ROOT_SUFFIX_RANDOM_INT`.
  * Else, the empty substring.
'''


CODE_GET_FUNC_PITH_VIOLATION = f''':
            {VAR_NAME_VIOLATION} = {ARG_NAME_GET_VIOLATION}(
                check_meta={ARG_NAME_CHECK_META},
                pith_name={{pith_name}},
                pith_value={VAR_NAME_PITH_ROOT},{{arg_random_int}}
            )
'''
'''
Code snippet suffixing all code type-checking the **root pith** (i.e., value of
the current parameter or return of a :func:`beartype.beartype`-decorated
callable) against the root type hint annotating that pith by either raising a
fatal exception or emitting a non-fatal warning.

This snippet expects to be formatted with these named interpolations:

* ``{arg_random_int}``, whose value is either:

  * If type-checking for the current type hint requires a pseudo-random integer,
    :data:`.CODE_HINT_ROOT_SUFFIX_RANDOM_INT`.
  * Else, the empty substring.
'''


CODE_GET_VIOLATION_RANDOM_INT = f'''
                random_int={VAR_NAME_RANDOM_INT},'''
'''
Code snippet passing the value of the random integer previously
generated for the current call to the exception-handling function call embedded
in the :data:`.CODE_HINT_ROOT_SUFFIX` snippet.
'''

# ....................{ CODE ~ violation                   }....................
CODE_RAISE_VIOLATION = f'''
            raise {VAR_NAME_VIOLATION}'''
'''
Code snippet raising the type-checking violation previously generated by the
:data:`.CODE_HINT_ROOT_SUFFIX` or
:data:`.PEP484_CODE_CHECK_NORETURN` code snippets as a fatal exception.
'''


CODE_WARN_VIOLATION = f'''
            {ARG_NAME_WARN}(str({VAR_NAME_VIOLATION}), type({VAR_NAME_VIOLATION}))'''
'''
Code snippet emitting the type-checking violation previously generated by the
:data:`.CODE_HINT_ROOT_SUFFIX` or
:data:`.PEP484_CODE_CHECK_NORETURN` code snippets as a non-fatal warning.
'''
